#include "ggnprivate.h"
#ifndef lint
static char rcsid[] ATTR_UNUSED = "$Id: common.c,v 1.1.1.1 1999/09/22 05:07:21 short Exp $";
#endif

#include <stdarg.h>
#include <sys/uio.h>
#include <netdb.h>
#include <sys/utsname.h>
#include <string.h>
#include <ctype.h>
#include <signal.h>
#include <errno.h>
#include <limits.h>
#include <sys/socket.h>

static int safesiga[]={SIGQUIT,SIGINT,SIGTERM,SIGIOT,SIGALRM,-1};
static char doexit;

char safegot;
char *localhostname,*localtaddr;
struct sockaddr_in localsaddr;
void (*closeconnp)(struct bufconn *);
#define closeconn(a) do { if (closeconnp) (*closeconnp)((a)); } while (0)
char (*fdsetopp)(int,unsigned);
#define fdsetop(a,b) do { if (fdsetopp) (*fdsetopp)((a),(b)); } while (0)
void (*presenterrp)(char *);
int cmdpar;
char *cmdpars[CMD_MAXPARS];
int *safesigs=safesiga;

void *remhead(struct list *lt)
{
void *res=lt->head;
	if (!res) return NULL;
	if (!(lt->head=*(void **)lt->head))
		lt->tail=(void *)&lt->head;
	return res;
}

void addtail(struct list *lt,void *nd)
{
	*(void **)lt->tail=nd;
	*(void **)nd=NULL;
	lt->tail=nd;
}

void remmid(struct list *lt,void *nd)
{
void **ltp=&lt->head;
	dbg(printf("remmid(nd=%p)\n",nd););
	while (*ltp!=nd) {
		dbg(printf("ltp=%p, *ltp=%p, nd=%p\n",ltp,*ltp,nd));
		ltp=(void **)*ltp;
		assert(!!ltp);
		}
	dbg(printf("ltp=%p, *ltp=%p, nd=%p\n",ltp,*ltp,nd));
	if (!(*ltp=*(void **)nd))
		lt->tail=(void *)ltp;
}

void writebuf(struct bufconn *conn,const char *buf,size_t len)
{
struct bufpart *dw;

	if (!len) return;
	chk(dw=malloc(sizeof(*dw)));
	chk(dw->buf=malloc(len));
	memcpy(dw->buf,buf,len);
	dw->done=0;
	dw->len=len;
	fdsetop(conn->fd,FSO_SET|FSO_WRITE);
	conn->writecnt++;
	addtail(&conn->delwrite,dw);
#ifdef DEBUGIO
	printf("%02u<<",conn->fd);
	fwrite(buf,len,1,stdout);
	if (buf[len-1]!='\n') puts("\\");
#endif
}

void chk(const void *ptr)
{
	if (!ptr) errout(_("Virtual memory exhausted"));
}

void errout(const char *fmt,...)
{
va_list ap;
char dos,back;
int err=errno;
char buf[LINE_MAX],*bufp;

	va_start(ap,fmt);
	if ((back=(*fmt=='&'))) fmt++;
	if (( dos=(*fmt=='^'))) fmt++;
	bufp=buf+VARVPRINTF(buf,fmt,ap);
	va_end(ap);
	if (dos) bufp+=VARPRINTFAT(buf,bufp," - %s",strerror(err));
	if (bufp+1<buf+sizeof(buf)) { *bufp++='!'; *bufp='\0'; }
	if (*presenterrp) (*presenterrp)(buf);
	else fprintf(stderr,"%s\n",buf);
	if (!back) exit(EXIT_FAILURE);
}

char saform_buf[SAFORM_BUF_SIZ];
const char *saform(const struct sockaddr_in *sap)
{
union {
	unsigned s4;
	unsigned char s1[4];
	} v;
	v.s4=ntohl(sap->sin_addr.s_addr);
	VARPRINTF5(saform_buf,"%u.%u.%u.%u:%u",v.s1[3],v.s1[2],v.s1[1],v.s1[0],
		ntohs(sap->sin_port));
	return(saform_buf);
}

const char *saformname(const struct sockaddr_in *sap)
{
struct hostent *he;
size_t l;
unsigned short port;
char *s;

	he=gethostbyaddr((char *)&sap->sin_addr.s_addr,4,AF_INET);
	if (!he || he->h_addrtype!=AF_INET || he->h_length!=4 || !he->h_name || !*he->h_name
		|| (l=strlen(he->h_name))>=sizeof(saform_buf)-16) {
		saform(sap);
		s=saform_buf+strlen(saform_buf);
		while (*--s!=':') assert(s>saform_buf);
		}
	else {
		memcpy(saform_buf,he->h_name,l);
		s=saform_buf+l;
		he=NULL;
		}
	port=ntohs(sap->sin_port);
	if (port==SERVICE_PORT) *s='\0';
	else if (!he) VARPRINTFAT(saform_buf,s,":%u",port);
	return(saform_buf);
}

void trywrite(struct bufconn *conn)
{
struct bufpart *dw;
ssize_t got;
#ifdef USE_WRITEV
struct iovec *iov,*iovb;
#endif
	dbg(printf("fd %d ready for write.\n",conn->fd));
#if 1
	if (!conn->writecnt) {
		dbg(printf("ALERT!!!ALERT!!!ALERT!!!ALERT!!!ALERT!!! Ready fd %p(%d) with EMPTY write queue!!!\n",conn,conn->fd));
		fdsetop(conn->fd,FSO_CLEAR|FSO_WRITE);
		return;
		}
#else
	assert(!!conn->writecnt);
#endif
#ifdef USE_WRITEV
	chk(iov=(void *)alloca(sizeof(*iov)*conn->writecnt));
	iovb=iov;
	for (dw=conn->delwrite.head;dw;dw=dw->next) {
		iov->iov_base=dw->buf+dw->done;
		iov->iov_len=dw->len;
		iov++;
		}
	if ((got=writev(conn->fd,iovb,conn->writecnt))<=0)
#else
	while ((dw=conn->delwrite.head)) {
		dbg(printf("write(%d,%p,%u).\n",conn->fd,dw->buf+dw->done,dw->len));
		if ((got=write(conn->fd,dw->buf+dw->done,dw->len))<=0)
#endif
		{
		if (!got&&errno==EAGAIN) {
			dbg(printf("Write failed as it would block\n"));
			return;
			}
		dbg(printf("Unspecific write error, closing...\n"));
		(*closeconnp)(conn);
		rem_bp(&conn->delwrite);
		return;
		}
#ifdef USE_WRITEV
	while (got&&(dw=conn->delwrite.head)) {
		if (got<dw->len) { dw->done+=got; dw->len-=got; return; }
		got-=dw->len;
#else
		dw->done+=got;
		if ((dw->len-=got)>0) return;
#endif
		remhead(&conn->delwrite);
		conn->writecnt--;
		dbg(printf("Removing write packet of fd %d.\n",conn->fd));
		freebufpart(dw);
		}
#ifdef USE_WRITEV
	assert(!got);
	assert(!conn->writecnt);
#endif
	dbg(printf("Removed last write packet of fd %d, clearing fdwr!\n",conn->fd));
	fdsetop(conn->fd,FSO_CLEAR|FSO_WRITE);
}

void rem_bp(struct list *bpl)
{
struct bufpart *bp;
	while ((bp=remhead(bpl))) freebufpart(bp);
}

struct hostent *resolve(const char *name,struct sockaddr_in *saddr)
{
static struct hostent *he;

	dbg(printf("gethostbyname(%s)\n",name));
	if (!(he=gethostbyname(name))) {
		errout("&^gethostbyname(%s)",name);
		return(NULL);
		}
	dbg(printf("gethostbyname(%s) ok\n",name));
	if (he->h_addrtype!=AF_INET||he->h_length!=sizeof(struct in_addr)||!*he->h_addr_list)
		errout(_("gethostbyname(%s) data not acceptable"),name);
	saddr->sin_family=AF_INET;
	memcpy(&saddr->sin_addr.s_addr,*he->h_addr_list,sizeof(struct in_addr));
	saddr->sin_port=htons(SERVICE_PORT);
	return(he);
}

void getlocalname(void)
{
char purename[256];
struct hostent *he;

	if (localhostname) return;
	if (gethostname(purename,sizeof(purename))) errout("^gethostname(,%d)",sizeof(purename));
	if (!(he=resolve(purename,&localsaddr))) errout(_("localhostname resolving"));
	chk(localhostname=strdup(he->h_name));
	chk(localtaddr=strdup(saform(&localsaddr)));
}

void linefuncize(struct bufconn *conn,struct bufpart *bp,void (*func)(char *,struct bufconn *))
{
char *s,*s2=NULL;
struct bufpart *bp2;
size_t siz,real,sofar;

	if (!bp) bp=conn->delread.head;
	if (bp==conn->delread.head) sofar=0;
	else if (bp==conn->delread.tail) {
		sofar=conn->delreadtot-bp->len;
		assert(!bp->done);
		}
	else {
		sofar=-(bp2=conn->delread.head)->done;
		for (sofar=0,bp2=conn->delread.head;bp2!=bp;bp2=bp2->next) sofar+=bp2->len;
		}
	while (conn->delreadtot) {
		assert(bp);
		if (!(s=memchr(bp->buf+bp->done,'\n',bp->len-bp->done))) {
			if ((bp=bp->next)) continue;
			return;
			}
		siz=sofar+s-(bp->buf+bp->done)+1;
		sofar=0;
		dbg(printf("siz=%u\n",siz));
		chk(s=malloc(siz));
		s2=s;
		while (s2-s<siz) {
			bp2=conn->delread.head;
			assert(!!bp2);
			memcpy(s2,bp2->buf+bp2->done,real=MIN(bp2->len-bp2->done,siz-(s2-s)));
			s2+=real;
			dbg(printf("Copied %u, now %u/%u\n",real,s2-s,siz));
			conn->delreadtot-=real;
			if ((bp2->done+=real)==bp2->len) {
				dbg(printf("Removing bp %p\n",bp2));
				remhead(&conn->delread);
				freebufpart(bp2);
				}
			}
		assert(s+siz==s2&&siz>=1&&s2[-1]=='\n');
		s2--;
		if (s2>s&&s2[-1]=='\r') s2--;
		*s2='\0';
		dbg(printf("Processing: \"%s\"\n",s));
		(*func)(s,conn);
		free(s);
		bp=conn->delread.head;
		}
}

void addfrag(struct bufconn *conn,char *buf,size_t got,void (*func)(char *,struct bufconn *))
{
struct bufpart *bp;

	dbg(printf("addfrag: conn=%p,buf=%p,got=%u,func=%p\n",conn,buf,got,func));
#ifdef DEBUGIO
	printf("%02u>>",conn->fd);
	fwrite(buf,got,1,stdout);
	if (!got || buf[got-1]!='\n') puts("\\");
#endif
	chk(bp=malloc(sizeof(*bp)));
	chk(bp->buf=malloc(got));
	bp->len=got;
	bp->done=0;
	memcpy(bp->buf,buf,got);
	dbg(printf("going to add bp=%p,buf=%p,len=%d to delread, got=%d\n",bp,bp->buf,bp->len,got));
	addtail(&conn->delread,bp);
	conn->delreadtot+=got;
	if (!func) {
		dbg(printf("Leaving fragment stored - prepared/in channel mode\n"));
		return;
		}
	dbg(printf("Extracting first line...\n"));
	linefuncize(conn,bp,func);
}

void parsepars(char *s)
{
char c,*d,quote;

	cmdpar=0;
	if (*s) for (*s++='\0';*s&&ISC(*s,space);s++);
	for (c=*s;c&&cmdpar<CMD_MAXPARS;cmdpar++) {
		quote=0;
		for (cmdpars[cmdpar]=d=s;*s&&(quote||(!quote&&*s!=','));s++) {
			if (*s=='"'||*s=='\'') {
				if (quote==*s) { quote=0; continue; }
				if (!quote) { quote=*s; continue; }
				}
			if (*s=='\\'&&s[1]) s++;
			*d++=*s;
			}
		c=*s++; *d='\0';
		}
}

static void safehandler(int signum)
{
	safegot=1;
	dbg(printf("In safehandler(%d)\n",signum));
	if (doexit) exit(EXIT_FAILURE);
}

void safeexit(char parexit)
{
int *ip=safesigs;
	doexit=parexit;
	while (*ip!=-1) {
		signal(*ip++,safehandler);
		dbg(printf("Registered safehandler on %d\n",ip[-1]));
		}
}
