#include "ggnprivate.h"
#ifndef lint
static char rcsid[] ATTR_UNUSED = "$Id: ggn.c,v 1.3 2000/09/08 09:51:02 short Exp $";
#endif

#include <netdb.h>
#include <string.h>
#include <limits.h>
#include <ctype.h>
#include <signal.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <fcntl.h>
#include <errno.h>
#include <fnmatch.h>
#include <stdarg.h>
#include <sys/socket.h>

struct conn {
	struct bufconn b;
	struct sockaddr_in *path;
	unsigned pathlast,pathdone;
	enum { CST_DEAD,CST_INIT,CST_PROXY,CST_HELLO,CST_COMMAND,CST_PLAYER,CST_MEMBER,CST_SAY,CST_START,CST_CHANNEL,CST_CHANWAIT,CST_CHANCNT,CST_REBROAD,CST_MASTERCLOSE } state;
	char *name;
	};

struct ggnconnect ggnconnpriv;
#define gconn ggnconnpriv

void (*notifyggnp)(ggnn_c,char **);
#define notifyggn(a,b) do { if (notifyggnp) (*notifyggnp)((a),(b)); } while (0)

#define B2CONN(bc) ((struct conn *)(bc))

static char linebuf[LINE_MAX];
static struct conn master,data,*cc;
static char isinit,mysigflag,noclosehack,datafail,retrymove,slept,rebroad;
static unsigned redirects,chancnt;
static time_t redirtime;
static pid_t daemonpid;
static char *wantplayer,*wantmember;
static char *  isplayer,*  ismember;
static char *wantdata,*wantmaster;
static int wantchan;
static struct list msgqueue;

struct msgnode {
	struct msgnode *next;
	enum { MSGN_SAY, MSGN_START } t;
	union {
		struct {
			char *msg;
			} say;
		struct {
			char *game,*server;
			int chan;
			} start;
		} u;
	};

static void writeccout(const char *msg,...)
{
va_list ap;
	va_start(ap,msg);
	writebuf(&cc->b,linebuf,VARVPRINTF(linebuf,msg,ap));
	va_end(ap);
}

static void mysigcatch(int signum)
{
	dbg(printf("mysigcatch()\n"));
	mysigflag=1;
}

static void killdaemon(void)
{
	if (!daemonpid) return;
	dbg(printf("Killing daemon on pid %d!\n",(int)daemonpid));
	kill(daemonpid,SIGTERM);
	alarm(DAEMON_KILL_TIME);
	waitpid(daemonpid,NULL,0);
	alarm(0);
	kill(daemonpid,SIGKILL);
	dbg(printf("Killed daemon.\n"));
}

static void force__exit(void)
{
	dbg(printf("Forcing exit...\n"));
	_exit(0);
}

static void closeconn(struct bufconn *bcn)
{
struct conn *cn=B2CONN(bcn);

	dbg(printf("Trying to close fd %d (master=%d)\n",bcn->fd,cn==&master));
	if (bcn->fd==-1) return;
	dbg(printf("Real close\n"));
	(*fdsetopp)(bcn->fd,FSO_CLEAR|FSO_READ|FSO_WRITE);
	rem_bp(&bcn->delwrite);
	rem_bp(&bcn->delread);
	bcn->delreadtot=0;
	cn->state=CST_DEAD;
	if (!noclosehack) (*fdsetopp)(bcn->fd,FSO_CLOSE);
	destroy(bcn->name);
	bcn->fd=-1;
	dbg(printf("Close finish\n"));
}

static void checkinit(void)
{
	if (isinit) return;
	closeconnp=closeconn;
	getlocalname();
	master.b.fd=-1;
	data.b.fd=-1;
	gconn.fd=-1;
	INIT_LIST(msgqueue);
	atexit(killdaemon);
	redirtime=time(NULL);
	isinit=1;
}

static void mastabort(void)
{
	dbg(printf("mastabort()\n"));
	if (master.b.fd!=-1) {
		if (master.state!=CST_MASTERCLOSE && master.state!=CST_DEAD) {
			writebuf(&master.b,"quit\n",5);
			master.state=CST_MASTERCLOSE;
			}
		}
	destroy(master.name);
	destroy(isplayer);
	destroy(ismember);
}

static void dataabort(void)
{
	dbg(printf("dataabort()\n"));
	if (data.b.fd!=-1) closeconn(&data.b);
	destroy(data.name);
	destroy(wantdata);
	wantchan=0;
	datafail=0;
}

void ggnabort(void)
{
	dbg(printf("ggnabort()\n"));
	checkinit();
	mastabort();
	closeconn(&master.b);
}

static void outmypath(void)
{
int i;

	for (i=cc->pathlast-1;i>=0;i--)
		writeccout("%s!",saform(cc->path+i));
}

static void connconn(char *dest);

static void action(void)
{
	if (cc->state==CST_INIT&&cc->pathdone<cc->pathlast) {
		dbg(printf("Wrote proxy %s\n",saform(cc->path+cc->pathdone+1)));
		writeccout("proxy %s\n",saform(cc->path+cc->pathdone+1));
		cc->state=CST_PROXY;
		return;
		}
	if (cc->pathdone<cc->pathlast) {
		dbg(printf("Netpath not yet complete...\n"));
		return;
		}
	if (cc->state==CST_INIT) {
		writeccout("hello \"");
		outmypath();
		writeccout("%s\"\n",localtaddr);
		cc->state=CST_HELLO;
		return;
		}
	if (cc->state==CST_CHANCNT) {
		if (chancnt||cc->b.writecnt) {
			dbg(printf("Something to do...\n"));
			return;
			}
		dbg(printf("Real start!!!\n"));
		gconn.fd=data.b.fd;
		mastabort();
		noclosehack=1;
		dataabort();
		noclosehack=0;
		retrymove=1;
		return;
		}
	if (cc==&master&&wantmaster&&master.state==CST_DEAD) {
char *newmast;
		newmast=wantmaster;
		wantmaster=NULL;
		connconn(newmast);
		if (wantdata) {
			cc=&data;
			connconn(wantdata);
			}
		/* if (!depth) */ notifyggn(GGNN_MASTER,&master.name);
		retrymove=1;
		return;
		}
	if (cc->state!=CST_COMMAND) return;
/* Data only commands block */
	if (cc==&data) {
		if (wantchan) {
			writeccout("channel %d\n",wantchan);
			cc->state=CST_CHANNEL;
			return;
			}
		return;
		}
	assert(cc==&master);
/* Master only commands follow */
	if (rebroad) {
		writeccout("rebroad\n");
		cc->state=CST_REBROAD;
		return;
		}
	dbg(printf("STR_EQ: wantplayer=\"%s\",isplayer=\"%s\"\n",wantplayer,isplayer));
	if (!STR_EQ(wantplayer,isplayer)) {
		dbg(printf("wantplayer correction\n"));
		writeccout("player \"%s\"\n",(wantplayer?wantplayer:""));
		cc->state=CST_PLAYER;
		return;
		}
	dbg(printf("STR_EQ: wantmember=\"%s\",ismember=\"%s\"\n",wantmember,ismember));
	if (isplayer&&!STR_EQ(wantmember,ismember)) {
		dbg(printf("wantmember correction\n"));
		writeccout("member \"%s\"\n",(wantmember?wantmember:""));
		cc->state=CST_MEMBER;
		return;
		}
	if (!LIST_EMPTY(msgqueue)) {
struct msgnode *mn=msgqueue.head;

		switch(mn->t) {
		case MSGN_SAY:
			writeccout("say \"%s\"\n",mn->u.say.msg);
			cc->state=CST_SAY;
			break;
		case MSGN_START:
			writeccout("start \"%s\",\"",mn->u.start.game);
			outmypath();
			writeccout(mn->u.start.server?mn->u.start.server:localtaddr);
			writeccout("\",%d\n",mn->u.start.chan);
			cc->state=CST_START;
			break;
		default: assert(0); }
		return;
		}
	if (data.b.fd!=-1&&data.state==CST_CHANWAIT) {
		mastabort();
		return;
		}
}

static void ggnconn(char *newmast);

static void connconn(char *dest)
{
int i;
char *s,*s2;
size_t namel=0;

	if (dest&&!*dest) dest=NULL;
#ifndef NDEBUG
	if (!cc->name) {
		dbg(printf("Old cc->name=\"%s\", new dest=\"%s\"\n",cc->name,dest));
		assert(!cc->name);
		}
#endif
	if (!dest) dest=localhostname;
	i=1;
	for (s=dest;*s;s++)
		if (*s=='!') i++;
	free(cc->name); chk(cc->name=strdup("")); namel=1;
	free(cc->path);
	dbg(printf("cc->name=\"%s\", %d components\n",cc->name,i));
	chk(cc->path=malloc(i*sizeof(*cc->path)));
	cc->pathlast=i-1;
	i=0;
	for (s=dest;*s;i++) {
char *name;
const char *sa;
size_t sal,nl,d;
		if (*s=='!') s++;
		for (s2=s;*s2&&*s2!='!'&&*s2!=':';s2++);
		chk(name=alloca(s2-s+1));
		memcpy(name,s,s2-s);
		name[s2-s]='\0';
		if (!resolve(name,cc->path+i)) {
			ggnconn(NULL);
			return;
			}
		if (*s2==':') {
char *s3;
long l=strtol(++s2,&s3,10);
			if (l<0||l>65535||(*s3&&*s3!='!')) {
				errout(_("&Invalid port value, starting with: \"%s\""),s2);
				return;
				}
			cc->path[i].sin_port=htons(l);
			}
		sal=strlen(sa=saformname(cc->path+i));
		if ((nl=strlen(cc->name)+1+sal)>=namel) {
			namel=MAX(MAX(namel*2,16),nl);
			chk(cc->name=realloc(cc->name,namel));
			if (!*cc->name) d=0;
			else { d=strlen(cc->name); cc->name[d++]='!'; }
			strcpy(cc->name+d,sa);
			}
		for (s=s2;*s&&*s!='!';s++);
		}
	dbg(printf("i==%d,cc->pathlast=%d\n",i,cc->pathlast));
	assert(i==cc->pathlast+1);
retrylocal:
	if ((cc->b.fd=socket(AF_INET,SOCK_STREAM,IPPROTO_TCP))==-1) {
		errout("&^socket(tcp)");
		return;
		}
	dbg(printf("Got pristine socket %d (master=%d)\n",cc->b.fd,cc==&master));
	if (connect(cc->b.fd,(struct sockaddr *)cc->path,sizeof(*cc->path))) {
int err=errno;
void (*origsig)(int);
int pipefds[2];

		if (close(cc->b.fd)) errout("&^close(sk)");
		cc->b.fd=-1;
		dbg(printf("err=%s,cc->name=%s,localhostname=%s,daemon=%d,slept=%d\n",
			strerror(errno),cc->name,localhostname,(int)daemonpid,slept));
		if (daemonpid) {
pid_t pid;
			errno=0;
			pid=waitpid(daemonpid,NULL,WNOHANG);
			if (pid==daemonpid||errno==ECHILD) daemonpid=slept=0;
			}
		dbg(printf("err=%s,cc->name=%s,localhostname=%s,daemon=%d,slept=%d\n",
			strerror(errno),cc->name,localhostname,(int)daemonpid,slept));
		if (err!=ECONNREFUSED|| !cc->name || strcmp(cc->name,localhostname)||(daemonpid&&slept)) {
			errno=err; errout("&^connect(%s)",saform(cc->path));
			ggnconn(NULL);
			return;
			}
		if (daemonpid&&!slept) {
			dbg(printf("SLEEPING!\n"));
			sleep(STARTUP_DELAY);
			slept=1;
			goto retrylocal;
			}

		dbg(printf("Starting daemon...\n"));
		if (pipe(pipefds)) errout(_("^pipe() failed"));
		if (pipefds[0]==pipefds[1] || pipefds[0]<0 || pipefds[1]<0) errout(_("pipe() sanity fail"));
		if (pipefds[0]>pipefds[1]) {
			int fd=pipefds[0];
			pipefds[0]=pipefds[1];
			pipefds[1]=fd;
			}
		mysigflag=0;
		origsig=signal(SIGNAL_NAME,mysigcatch);
		if (!(daemonpid=fork())) {
char *argv[2];
int maxfd=pipefds[1]-1;
#if 0
			chk(*argv=alloca(strlen(pname)+strlen(DAEMON_FMT)-2+1));
			sprintf(*argv,DAEMON_FMT,pname);
#endif
			if (maxfd>10000) maxfd=1024;
			while (maxfd
#if defined(DEBUG)
				>STDERR_FILENO
#else
				>=0
#endif
				) close(maxfd--);
			presenterrp=NULL; /* Is it okay? */
			atexit(force__exit);
			argv[0]="fake (ggnd)";
			argv[1]=NULL;
			exit(ggnd_main(1,argv,pipefds[1]));
			}
		dbg(printf("Entering wait being the master of that beast...\n"));
		sleep(DAEMONWAITSEC);
		dbg(printf("Beast has signalled us (wait returned)!\n"));
		if (!mysigflag) {
			errout(_("&^Waiting for daemon start (%d seconds)"),DAEMONWAITSEC);
			ggnabort();
			}
		signal(SIGNAL_NAME,origsig);
		dbg(printf("Daemon forked.\n"));
		goto retrylocal;
		}
	(*fdsetopp)(cc->b.fd,FSO_SET|FSO_READ);
	cc->pathdone=0;
	INIT_LIST(cc->b.delwrite);
	INIT_LIST(cc->b.delread);
	cc->b.delreadtot=0;
	cc->b.writecnt=0;
	chk(cc->b.name=strdup(saformname(cc->path)));
	if (fcntl(cc->b.fd,F_SETFL,O_NONBLOCK)) {
    errout("&^fcntl(s,O_NONBLOCK)");
		ggnconn(NULL);
		}
	cc->state=CST_INIT;
	action();
}

static void ggnconn(char *newmast)
{
time_t rtime;

	if (cc==&data) {
		dataabort();
		datafail=1;
		return;
		}
	assert(cc==&master);
	if (newmast&&!*newmast) newmast=NULL;
	rtime=redirtime;
	redirtime=time(NULL);
	dbg(printf("dtime=%lu,redirects=%d,newmast=%s,master=%s,localname=%s\n",
		redirtime-rtime,redirects,newmast,master.name,localhostname));
	if (redirtime-rtime<MIN_OKAY_REDIRECT) redirects++;
	else redirects=0;
	if (master.name&&!strcmp(master.name,localhostname)&&redirects>=MAX_REDIRECTS)
		return;
	if (newmast&&(redirects>=MAX_REDIRECTS||!strcmp(newmast,localhostname)))
		newmast=NULL;
	rebroad=!newmast;
	mastabort();
	free(wantmaster); chk(wantmaster=strdup(newmast?newmast:localhostname));
	retrymove=1;
}

static void readresp(char *buf,struct bufconn *trash)
{
int i,code;
char *s,*s2,quoted=0;
static const struct {
	char parn;
	ggnn_c msgt;
	} tab3xx[]={
/*{1,GGNN_MASTER   },*/ /* 300=Current best network master */
  {2,GGNN_NEWPLAYER},   /* 301=New player connected        */
  {1,GGNN_DISPLAYER},   /* 302=Player disconnected         */
  {2,GGNN_MEMBER   },   /* 303=Player is member            */
  {4,GGNN_STARTGAME},   /* 304=Starting netgame            */
  {2,GGNN_MESSAGE  },   /* 305=Message from %s             */
};
#define TAB3XX_SIZ (sizeof(tab3xx)/sizeof(*tab3xx))
#define ARG2L(l,n) do { (l)=strtol(cmdpars[(n)],&s2,0); if ((l)<=0||(l)>=INT_MAX||*s2) goto bad; } while (0)

	if (cc->state==CST_CHANCNT) {
		assert(cc==&data);
		/* FIXME: "buf" should be checked for "#:" format */
		if (chancnt) {
char *err;
long l;
			l=strtol(buf,&err,0);
			if (!err || *err!=':') {
				errout(_("&No from-field in channel-names setup on %s: \"%s\""),cc->b.name,buf);
				ggnconn(NULL); return; }
			if (l<1 || l>gconn.tot || gconn.names[l-1]) {
				errout(_("&Invalid %ld from-field in channel-names setup on %s: \"%s\""),l,cc->b.name,buf);
				ggnconn(NULL); return; }
			chk(gconn.names[l-1]=strdup(err+1));
			dbg(printf("In chancnt, countdown=%u\n",chancnt));
			chancnt--;
			}
		return;
		}
	while (*buf=='>') buf++;
	for (i=0;i<4;i++)
		if ((i<3&&!ISC(buf[i],digit))||(i==3&&buf[i]!=' ')) {
			errout(_("&Unprocessable response on %s: \"%s\""),cc->b.name,buf);
			ggnconn(NULL); return; }
	if (*buf=='0') return; /* Informational */
	if (*buf=='3'&&(cc->pathdone<cc->pathlast||cc!=&master)) /* Notifications */
		return;
		
	code=atoi(buf);
	for (s=buf+strlen(buf)-1;s>=buf;s--) {
		if (!quoted&&(*s=='"'||*s=='\'')) quoted=*s;
		else if (quoted==*s) quoted=0;
		else if (!quoted&&*s==':') break;
		}
	if (s<buf) cmdpar=0;
	else {
		assert(*s==':');
		chk(s2=alloca(strlen(s)+1));
		strcpy(s2,s);
		parsepars(s2);
		}
	dbg(printf("Got %d pars\n",cmdpar));

	if (code==210&&cc->state==CST_PLAYER&&cmdpar==1) { /* Player name already exists, use another */
		if (wantplayer&&!strcmp(wantplayer,cmdpars[0])) {
			destroy(wantplayer);
			destroy(wantmember);
			}
		cc->state=CST_COMMAND;
		notifyggn(GGNN_ERRPLAYER,cmdpars);
		retrymove=1;
		return;
		}
	
	if (*buf=='2'||*buf=='4') { /* Error/Fatal */
		errout(_("&Failure on %s: %s"),cc->b.name,buf);
		ggnconn(NULL);
		return;
		}

	if (code==100&&cc->state==CST_HELLO) { /* Identity accepted */
		cc->state=CST_COMMAND;
		return;
		}
	if (code==101&&cc->state==CST_MASTERCLOSE) { /* Closing connection... */
		closeconn(&master.b);
		cc->state=CST_DEAD;
		return;
		}
	if ((code==109||code==110)&&cc->state==CST_REBROAD) {
		rebroad=0;
		cc->state=CST_COMMAND;
		return;
		}
	if (code==103&&cc->state==CST_PROXY&&cc->pathdone<cc->pathlast) { /* Proxying to %s and leaving... */
		free(cc->name);
		chk(cc->name=strdup(saform(cc->path+ (++cc->pathdone))));
		cc->state=CST_INIT;
		dbg(printf("Accepted proxy response, now pathdone=%d, pathlast=%d\n",cc->pathdone,cc->pathlast));
		return;
		}
	if (code==105&&cc->state==CST_CHANNEL&&cmdpar==1) { /* Channel created, waiting for %d members */
long chan;
struct msgnode *mn;

		assert(cc==&data);
		
		ARG2L(chan,0);
		wantchan=0;

		if (!wantmember) {
			dataabort();
			return;
			}
		chk(mn=malloc(sizeof(*mn)));
		mn->t=MSGN_START;
		chk(mn->u.start.game=strdup(wantmember));
		chk(mn->u.start.server=strdup(cc->name));
		mn->u.start.chan=chan;
		addtail(&msgqueue,mn);
		cc->state=CST_CHANWAIT;
		retrymove=1;
		return;
		}
	if (code==115&&cc->state==CST_CHANNEL) { /* Joined to channel %d, waiting still for %u members */
		if (!wantmember) dataabort();
		else cc->state=CST_CHANWAIT;
		return;
		}
	dbg(printf("Check: code=%d,state=%d,cmdpar=%d\n",code,cc->state,cmdpar));
	if (code==106&&cc->state==CST_CHANWAIT&&cmdpar==2) { /* Joined to channel %d, setting up channel mode, leaving */
long cur,tot;
char *s=wantplayer?wantplayer:"<unknown>";

		assert(cc==&data);

		ARG2L(cur,0); ARG2L(tot,1);
		gconn.cur=cur;
		while (gconn.tot--) free(gconn.names[gconn.tot]);
		free(gconn.names);
		gconn.tot=tot;
		chk(gconn.names=malloc(sizeof(*gconn.names)*gconn.tot));
		memset(gconn.names,0,sizeof(*gconn.names)*gconn.tot);
		chk(gconn.names[gconn.cur-1]=strdup(s));

		writeccout("%s\n",s);
		chancnt=tot-1;

		cc->state=CST_CHANCNT;
		retrymove=1;
		return;
		}
	if (code==111&&cc->state==CST_PLAYER&&cmdpar==1) { /* Player identity accepted */
		assert(cc==&master);

		destroy(isplayer);
		destroy(ismember);
		if (*cmdpars[0]) chk(isplayer=strdup(cmdpars[0]));
		cc->state=CST_COMMAND;
		notifyggn(GGNN_OKPLAYER,cmdpars);
		retrymove=1;
		return;
		}
	if (code==112&&cc->state==CST_MEMBER&&cmdpar==2) { /* Player considered as member */
		assert(cc==&master);

		destroy(isplayer);
		destroy(ismember);
		if (*cmdpars[0]) chk(isplayer=strdup(cmdpars[0]));
		if (*cmdpars[1]) chk(ismember=strdup(cmdpars[1]));
		cc->state=CST_COMMAND;
		notifyggn(GGNN_OKMEMBER,cmdpars+1);
		retrymove=1;
		return;
		}
	if (code==113&&cc->state==CST_SAY) { /* Message accepted for delivery */
struct msgnode *mn;

		assert(cc==&master);

		if (!msgqueue.head||((struct msgnode *)msgqueue.head)->t!=MSGN_SAY) goto bad;
		mn=remhead(&msgqueue);
		free(mn->u.say.msg);
		free(mn);
		cc->state=CST_COMMAND;
		return;
		}
	if (code==114&&cc->state==CST_START) { /* Netgame start ok */
struct msgnode *mn;

		assert(cc==&master);

		if (!msgqueue.head||((struct msgnode *)msgqueue.head)->t!=MSGN_START) goto bad;
		mn=remhead(&msgqueue);
		free(mn->u.start.game);
		free(mn->u.start.server);
		free(mn);
		cc->state=CST_COMMAND;
		return;
		}
	if (code==300&&cmdpar==1) { /* Current best network master */
		assert(cc==&master);

		dbg(printf("cmdpars[0]=%s\n",cmdpars[0]));
		for (s=master.name+strlen(master.name)-1;s>=master.name&&*s!='!';s--);
		s++;
		chk(s2=alloca(s-master.name+strlen(cmdpars[0])+1));
		memcpy(s2,master.name,s-master.name);
		strcpy(s2+(s-master.name),cmdpars[0]);
		dbg(printf("ggnconn(%s)\n",s2));
		ggnconn(s2);
		return;
		}
	if (code>=301&&(code-301)<TAB3XX_SIZ&&tab3xx[code-301].parn==cmdpar) {
		assert(cc==&master);

		notifyggn(tab3xx[code-301].msgt,cmdpars);
		retrymove=1;

		/* FIXME: To prevent (n+1) to (n) expected game fights, we prefer
		 * ismember to wantmember. Unfortunately there is another small
		 * but worse race window anyway - (n-1) to (n). */
		dbg(printf("Check: code=%d, cmdpars[1]=\"%s\", ismember=\"%s\"\n",code,cmdpars[1],ismember));
		if (code==304&&ismember&&!strcmp(cmdpars[1],ismember)) { /* Starting netgame */
long l;
			l=strtol(cmdpars[3],&s2,0);
			if (l<=0||l>=INT_MAX||*s2) goto bad;

			dataabort();
			chk(wantdata=strdup(cmdpars[2]));
			wantchan=l;
			}
		return;
		}

bad:
	errout(_("&Unexpected response on %s: %s"),cc->b.name,buf);
	ggnconn(NULL);
}

static void connmove(struct conn *conn)
{
int got; /* size_t is unsigned! */

	cc=conn;
	if (conn->b.fd!=-1&&(*fdsetopp)(conn->b.fd,FSO_WRITE)) trywrite(&conn->b);
	if (conn->b.fd!=-1&&(*fdsetopp)(conn->b.fd,FSO_READ)&&!(conn->state==CST_CHANCNT&&!chancnt)) {
		dbg(printf("tryread %d\n",conn->b.fd));
		if ((got=read(conn->b.fd,linebuf,sizeof(linebuf)))<=0) {
			if (!got)
				dbg(printf("Reading returned 0 - Closing connection\n"));
			else {
				if (errno==EAGAIN)
					goto action;
				dbg(printf("Read error\n"));
				}
			if (got) errout("&^read(%s)",conn->name);
			/* Our socket is already closed by peer! - hard close! */
			closeconn(&conn->b);
			ggnconn(NULL);
			goto action;
			}
		addfrag(&conn->b,linebuf,got,readresp);	
		}
action:
	action();
	dbg(printf("state=%d (master=%d)\n",cc->state,cc==&master));
}

static inline int dataokay(void)
{
	if (datafail||STR_EQ(wantdata,data.name)) return(1);
	if (wantdata&&data.name&&!*wantdata&&!strcmp(data.name,localhostname)) return(1);
	return(0);
}

struct ggnconnect *ggnmove(void)
{
	checkinit();
	do {
		retrymove=0;
		connmove(&master);
		connmove(&data);
		if (gconn.fd!=-1) {
			gconn.r=GGNC_OKAY;
			dbg(printf("Terminating OKAY ggnmove()\n"));
			return(&gconn);
			}
		if (!dataokay()) {
			dbg(printf("wantdata=\"%s\",data.name=\"%s\" - conflict!!! - reopening\n",wantdata,data.name));
			closeconn(&data.b);
			if (wantdata) {
				cc=&data;
				connconn(wantdata);
				}
			dbg(printf("datafail=%d,wantdata=\"%s\",data.name=\"%s\"\n",datafail,wantdata,data.name));
			assert(dataokay());
			}
		dbg(printf("retrymove=%d\n",retrymove));
		} while (retrymove);
	gconn.r=GGNC_ITERATE;
	dbg(printf("Terminating ITERATE ggnmove()\n"));
	return(&gconn);
}

void ggnmember(char *player,char *member)
{
	dbg(printf("ggnmember(\"%s\",\"%s\")\n",player,member));
	destroy(wantplayer);
	if (player&&!*player) player=NULL;
	if (player) chk(wantplayer=strdup(player));
	destroy(wantmember);
	if (member&&!*member) member=NULL;
	if (member&&wantplayer) chk(wantmember=strdup(member));
	retrymove=1;
}

void ggnsendmsg(char *msg)
{
struct msgnode *mn;

	assert(!!msg);
	checkinit();
	chk(mn=malloc(sizeof(*mn)));
	mn->t=MSGN_SAY;
	chk(mn->u.say.msg=strdup(msg));
	addtail(&msgqueue,mn);
	retrymove=1;
}

void ggnbegin(char *newmast)
{
	checkinit();
	gconn.fd=-1;
	dataabort();
	redirects=0;
	cc=&master;
	ggnconn(newmast);
}

void ggnstart(char *server,unsigned others)
{
	dataabort();
	if (others) /* else abort */
		chk(wantdata=strdup(server?server:""));
	wantchan=-others;
	retrymove=1;
}

static ggnboxfunc * const boxa[]={
#ifdef HAVE_GTK
ggnboxgtk,
#endif
ggnboxcon,
NULL};

struct ggnconnect *ggnbox(char *title,char **statep,int *argcp,char ***argvp,char *filter)
{
ggnboxfunc * const *bi;

	for (bi=boxa;*bi;bi++) {
struct ggnconnect *gc;
		if (filter) {
int minusone=-1;
			if (fnmatch(filter,(char *)(*bi)(NULL,NULL,&minusone,NULL),FNM_PATHNAME)) continue;
			}
		gc=(*bi)(title,statep,argcp,argvp);
		if (gc->r!=GGNC_FAIL) return (gc);
		}
	ggnconnpriv.r=GGNC_FAIL;
	return(&ggnconnpriv);
}

static char ** const stateppap[]={&master.name,&wantplayer,&wantmember,NULL};
static char *upmaster,*upplayer,*upmember;
static char ** const stateppau[]={&upmaster,&upplayer,&upmember,NULL};

char *ggnpackstate(void)
{
char ** const *sss,*s,*s2;
size_t l=0;

	dbg(printf("ggnpackstate()...\n"));
	for (sss=stateppap;*sss;sss++)
		l+=1+(**sss?strlen(**sss):strlen(GGNPACKNULL));
	chk(s=malloc(l));
	s2=s;
	for (sss=stateppap;*sss;sss++) {
		if (**sss) {
			memcpy(s,**sss,(l=strlen(**sss)));
			s+=l;
			}
		else {
			memcpy(s,GGNPACKNULL,strlen(GGNPACKNULL));
			s+=strlen(GGNPACKNULL);
			}
		*s++=GGNPACKCHAR;
		}
	s[-1]='\0';
	dbg(printf("ggnpackstate() returning\n"));
	return(s2);
}

void ggnunpackstate(char *state)
{
char ** const *sss,*s;

	dbg(printf("ggnunpackstate()...\n"));
	if (!state) state=GGN_DEFAULT_STATE;
	for (sss=stateppau;*sss;sss++) {
		assert(!**sss && state);
		if (!(s=strchr(state,GGNPACKCHAR))) s=state+strlen(state);
		if (!(s-state==strlen(GGNPACKNULL)&&!memcmp(GGNPACKNULL,state,strlen(GGNPACKNULL)))) {
			chk(**sss=malloc(s-state+1));
			memcpy(**sss,state,s-state);
			(**sss)[s-state]='\0';
			}

		if (*s) {
			assert(*s==GGNPACKCHAR);
			s++;
			}
		state=s;
		}
	dbg(printf("Depacked (\"%s\",\"%s\",\"%s\"), setting...\n",upmaster,upplayer,upmember));
	ggnbegin(upmaster);
	ggnmember(upplayer,upmember);
	destroy(upmaster);
	destroy(upplayer);
	destroy(upmember);
	dbg(printf("ggnunpackstate() returning\n"));
}

const char *ggnnotifymsg(ggnn_c code)
{
static const char *tab[]={
N_("Connected to master \"%s\"."),
N_("New player \"%s\" appeared from host \"%s\"."),
N_("Player \"%s\" has disconnected."),
N_("Player \"%s\" became member of game \"%s\"."),
N_("Player \"%s\" started game \"%s\" on server \"%s\", channel %s."),
N_("Message from \"%s\": \"%s\""),
N_("Player name \"%s\" registered."),
N_("Player \"%s\" already exists! Player not registered."),
N_("Now became member of \"%s\"."),
};

	if (code<GGNN_MASTER||code>GGNN_OKMEMBER)
		return(_("Unhandled notification code!"));
	return(_(tab[code-GGNN_MASTER]));
}
