#include "ggnprivate.h"
#ifndef lint
static char rcsid[] ATTR_UNUSED = "$Id: ggnboxcon.c,v 1.1.1.1 1999/09/22 05:07:21 short Exp $";
#endif

#include <stdio.h>
#include <stdlib.h>
#include <signal.h>
#include <string.h>
#include <limits.h>
#include <fcntl.h>
#include <errno.h>

static fd_set fdrd,fdwr,fdrdx,fdwrx;
static char linebuf[LINE_MAX],*pars,doexit;
static size_t filled;
static char **mstatep;
#define show_prompt() do { putchar('>'); fflush(stdout); } while (0)
#define upd_state() do { free(*mstatep); *mstatep=ggnpackstate(); } while (0)

FDSETOP(fdrd,fdrdx,fdwr,fdwrx)

static void presenterr(char *msg)
{
	printf(_("GGN PresentErr: \"%s\"\n"),msg);
}

static void notifyggn(ggnn_c code,char **sp)
{
	printf(_("GGN Notify: "));
	printf(ggnnotifymsg(code),sp[0],sp[1],sp[2],sp[3]);
	putchar('\n');
#ifdef DEBUG
	if (code==GGNN_ERRPLAYER) {
char *s=alloca(strlen(*sp)+2);
		strcpy(s,*sp);
		strcat(s,"+");
		ggnmember(s,"MEMBER");
		}
#endif
}

static void fn_help(void)
{
	printf(_("\
? - this help\n\
b - connect to new master[1]\n\
a - abort all connections (incl. master!)\n\
m - we are player[1] in game[2]\n\
s - start on server[1] (\"\" for local) with (n+1)[2] players\n\
M - send broadcast message[1]\n\
q - Quit\n\
"));
}

static void fn_quit(void)
{
	if (mstatep) upd_state();
	printf("ggnabort()\n");
	ggnabort();
	printf(_("Quitting command-line client...\n"));
	doexit=1;
}

static void fn_begin(void)
{
	printf("ggnbegin(\"%s\")\n",pars);
	ggnbegin(pars);
}

static void fn_abort(void)
{
	printf("ggnabort()\n");
	ggnabort();
}

static void fn_member(void)
{
char *s;

	if (!(s=strchr(pars,','))) {
		printf(_("ggnmember() needs two parametes!\n"));
		return;
		}
	*s='\0';
	printf("ggnmember(\"%s\",\"%s\")\n",pars,s+1);
	ggnmember(pars,s+1);
	*s=',';
}

static void fn_start(void)
{
char *s;
int i;

	if (!(s=strchr(pars,','))) {
		printf(_("ggnstart() needs two parametes!\n"));
		return;
		}
	*s='\0';
	i=atoi(s+1);
	printf("ggnstart(\"%s\",%d)\n",pars,i);
	ggnstart(pars,i);
	*s=',';
}

static void fn_sendmsg(void)
{
	printf("ggnsendmsg(\"%s\")\n",pars);
	ggnsendmsg(pars);
}

static const struct {
	char c;
	void (*f)(void);
	} cmdtab[]={
{'?',fn_help},
{'b',fn_begin},
{'a',fn_abort},
{'m',fn_member},
{'s',fn_start},
{'M',fn_sendmsg},
{'q',fn_quit},
};
#define CMDTAB_SIZ (sizeof(cmdtab)/sizeof(*cmdtab))

struct ggnconnect *ggnboxcon(char *title,char **statep,int *argcp,char ***argvp)
{
int got;
struct ggnconnect *gc;

	GGNBOX_ID("console/con");
	safeexit(1);
	fdsetopp=fdsetop;
	notifyggnp=notifyggn;
	presenterrp=presenterr;
	FD_ZERO(&fdrd); FD_ZERO(&fdwr);
	doexit=filled=0;
	mstatep=statep;

	FD_SET(STDIN_FILENO,&fdrd);
  if (fcntl(STDIN_FILENO,F_SETFL,O_NONBLOCK))
	    errout("&^fcntl(s,O_NONBLOCK)");
	printf(_("%s\nCommand-line client v%s. Use '?' for help.\n\n"),title,VERSION);
	
	ggnunpackstate(*statep);

	show_prompt();

	do {
		fdrdx=fdrd;
		fdwrx=fdwr;
		select(FD_SETSIZE,&fdrdx,&fdwrx,NULL,NULL);
		dbg(printf("FD_ISSET=%d\n",(int)(FD_ISSET(STDIN_FILENO,&fdrdx))));
		if (FD_ISSET(STDIN_FILENO,&fdrdx)) {
			while ((got=read(STDIN_FILENO,linebuf+filled,sizeof(linebuf)-filled))>0) {
char *s;
int i;

				dbg(printf("got=%d\n",got));
				filled+=got;
				while ((s=memchr(linebuf,'\n',filled))) {
					*s='\0';
					if (*linebuf) {
						pars=linebuf+1;
						for (i=0;i<CMDTAB_SIZ;i++)
							if (*linebuf==cmdtab[i].c) {
								(*cmdtab[i].f)();
								if (doexit) {
									ggnconnpriv.r=GGNC_QUIT;
									return(&ggnconnpriv);
									}
								break;
								}
						if (i==CMDTAB_SIZ)
							printf(_("Unrecognized command: `%c' - use ? for help\n"),*linebuf);
						}
					memcpy(linebuf,s+1,(filled-=(s+1-linebuf)));
					show_prompt();
					}
				}
			dbg(printf("got=%d, err=%s\n",got,strerror(errno)));
			if (!got) {
				fn_quit();
				if (doexit) {
					ggnconnpriv.r=GGNC_QUIT;
					return(&ggnconnpriv);
					}
				}
			}
		gc=ggnmove();
		} while (gc->r==GGNC_ITERATE);
	if (gc->r!=GGNC_OKAY) return(gc);
	printf(_("Connected, we are player %d out of %d total.\n"),gc->cur,gc->tot);
	upd_state();
	return(gc);
}
