#ifdef HAVE_CONFIG_H
#include "config.h"
#endif
#ifndef lint
static char rcsid[] ATTR_UNUSED = "$Id: billggn.c,v 1.1.1.1 1999/09/22 05:07:21 short Exp $";
#endif

#include <fcntl.h>
#include <errno.h>
#include <stdarg.h>
#include <string.h>
#include <pwd.h>
#include <sys/types.h>
#include <stdio.h>
#include <stdlib.h>
#include <assert.h>
#ifdef HAVE_GNOME
#include <gnome.h>
#endif

#include "billggn.h"
#include "game.h"
#include "ui.h"

GGN ggn;

#define MINBUF_SIZE 256

void ggn_blocking(char state)
{
	if (!ggn_active()) return;
  	fcntl(ggn.gc->fd,F_SETFL,(state?0:O_NONBLOCK));
}

void ggn_menu(void)
{
int argc=2;
char *argv[]={"xBill","--gtk-is-initialized",NULL};
char **argvp=argv;
static char inside=0;

	if (inside) return;
	ggn_disconn();
	inside=1;
	ggn.gc=ggnbox("xBill",&ggn.state,&argc,&argvp,"x11/*");
	inside=0;
	if (ggn.gc && ggn.gc->r!=GGNC_OKAY) ggn.gc=NULL;
#ifdef HAVE_GNOME
	gnome_config_set_string(PACKAGE "/ggn/state",ggn.state);
	gnome_config_sync();
#endif
	if (!ggn_active()) {
		ui_set_statusbar(_("Network setup connection cancelled."));
		return;
		}
	ui_set_statusbar(_("Network setup completed, now internetworking..."));
	ggn.fout=fdopen(ggn.gc->fd,"w");
	setlinebuf(ggn.fout);
	chk(ggn.gotplayers=(char *)realloc(ggn.gotplayers,ggn.gc->tot*sizeof(*ggn.gotplayers)));
	ui_set_ggndisconnbutton(1);
	ggn.inputid=gdk_input_add(ggn.gc->fd,GDK_INPUT_READ,(GdkInputFunction)ggn_fdreadf,NULL);
	ggn_resetmove();
	game_clear();
	ui_refresh();
}

void ggn_resetmove(void)
{
	if (!ggn_active()) return;
	memset(ggn.gotplayers,0,ggn.gc->tot);
	ggn.remaining=ggn.gc->tot-1;
}

void ggn_disconn(void)
{
	if (!ggn_active()) return;
	ui_set_statusbar(_("Network connection has been closed."));
	ui_set_ggndisconnbutton(0);
	gdk_input_remove(ggn.inputid);
	ggn_blocking(0);
	fclose(ggn.fout);
	close(ggn.gc->fd); /* probably redundant */
	ggn.gc=NULL;
	ggn.me=0;
}

void ggn_setup(void)
{
struct passwd *pwd;

	ui_set_ggndisconnbutton(0);
#ifdef HAVE_GNOME
	ggn.state=gnome_config_get_string(PACKAGE "/ggn/state");
	if (ggn.state && !*ggn.state) { g_free(ggn.state); ggn.state=NULL; }
	if (ggn.state) return;
#endif
	pwd=getpwuid(getuid());
	if (!pwd || !pwd->pw_name) return;
	chk(ggn.state=malloc(strlen(pwd->pw_name)+32));
	sprintf(ggn.state,GGNPACKNULL"%c%s%c%sGame",GGNPACKCHAR,pwd->pw_name,GGNPACKCHAR,pwd->pw_name);

}

/* FIXME: better random generator! */
void ggn_Gsrand(unsigned seed)
{
	ggn.randval=seed;
}

int ggn_Grand(void)
{
int r=ggn.randval;
	ggn.randval=ggn.randval*0xDEECE66D+0xB;
	if (r>=0) return(r);
	else return(-r);
}

void ggn_fdreadf(void)
{
int got;
char *s,*d;

	if (!ggn_active()) return;
	if (setjmp(ggn.jmp)) { ggn_disconn(); return; }
	for (;;) {
		if (ggn.buf.size==ggn.buf.len) {
			if ((ggn.buf.size*=2)<MINBUF_SIZE) ggn.buf.size=MINBUF_SIZE;
			chk(ggn.buf.data=(char *)realloc(ggn.buf.data,ggn.buf.size));
			}
		assert(ggn.buf.size>ggn.buf.len && ggn.buf.data);
		ggn_blocking(0);
		errno=0; got=read(ggn.gc->fd,ggn.buf.data+ggn.buf.len,ggn.buf.size-ggn.buf.len);
		assert(got>=-1);
		if (got==-1 && errno==EAGAIN) return;
		if (got<=0) { d(("NETERROR: got=%d,err=%s\n",got,strerror(errno))); ggn_disconn(); return; }
		ggn.buf.len+=got;
		while ((s=(char *)memchr(ggn.buf.data,'\n',ggn.buf.len))) {
			d=s; s++;
			if (d>ggn.buf.data&&d[-1]=='\r') d--;
			*d='\0';
			ggn.me=1;
			ggn.buf.s=ggn.buf.data;
			ggn_gotline();
			ggn.me=0;
			ggn.buf.len-=(s-ggn.buf.data);
			memmove(ggn.buf.data,s,ggn.buf.len);
			}
		}
}

int ggn_getint(void)
{
int r;
char *err;
	r=strtol(ggn.buf.s,&err,0);
	if (!err || *err!=',') { d(("LONGJMP: *err=%c\n",*err)); longjmp(ggn.jmp,1); }
	ggn.buf.s=err+1;
	return(r);
}

void ggn_gotline(void)
{
char *err,cmd;
long from;
	assert(ggn_active());
	d(("gotline: %s\n",ggn.buf.s));
	from=strtol(ggn.buf.s,&err,0);
	if (!err || *err!=':' || from<1 || from>ggn.gc->tot || from==ggn.gc->cur) {
		d(("LONGJMP: *err=%c,from=%ld,ggn.gc->tot=%d,ggn.gc->cur=%d\n",
			*err,from,ggn.gc->tot,ggn.gc->cur));
		longjmp(ggn.jmp,1);
		}
	ggn.buf.s=err+1;
	while ((cmd=*ggn.buf.s++)) switch (cmd) {
		case 'S':
		case 's':
			ggn_Gsrand(ggn_getint());
			game_setup_level((cmd=='S'),ggn_getint());
			return;
		case 'p': {
int x,y;
			x=ggn_getint(); y=ggn_getint();
			game_button_press(x,y,ggn.gc->names[from-1]);
			} break;
		case 'r': {
int x,y;
			x=ggn_getint(); y=ggn_getint();
			game_button_release(x,y,ggn.gc->names[from-1]);
			} break;
		default: d(("LONGJMP: cmd=%c\n",cmd)); longjmp(ggn.jmp,1);
		}
	if (!ggn.gotplayers[from-1]) {
		ggn.gotplayers[from-1]=1;
		assert(ggn.remaining>0);
		if (!--ggn.remaining) game_update(1);
		}
}

void ggn_netprintf(const char *fmt,...)
{
va_list ap;
	dbg(if (ggn_active()) printf("%d: ggn_netprintf(\"%s\"), serving=%d\n",ggn.gc->cur,fmt,ggn_serving()));
	if (!ggn_serving()) return;
	ggn_blocking(1);
	va_start(ap,fmt);
	vfprintf(ggn.fout,fmt,ap);
	va_end(ap);
}
