#ifdef HAVE_CONFIG_H
#include "config.h"
#endif
#ifndef lint
static char rcsid[] ATTR_UNUSED = "$Id: cable.c,v 1.1.1.1 1999/09/22 05:07:21 short Exp $";
#endif

#include "cable.h"
#include "network.h"
#include "library.h"
#include "ui.h"
#include "spark.h"
#include "game.h"

void cable_setup(Cable *C) {
	C->c1 = game_RAND(0, net.units-1);
	do {
		C->c2 = game_RAND(0, net.units-1);
	} while (C->c2 == C->c1);
	C->active = C->index = 0;
	C->delay = spark_delay(game.level);
	C->x1 = net.computers[C->c1].x + net.width/3;
	C->x2 = net.computers[C->c2].x + net.width/3;
	C->y1 = net.computers[C->c1].y + net.height/2;
	C->y2 = net.computers[C->c2].y + net.height/2;
}

void cable_update (Cable *C) {
	if (C->active) {
		if ((net.computers[C->c1].os == OS_WINGDOWS) ==
			(net.computers[C->c2].os == OS_WINGDOWS))
				C->active=0;
		else if (net.computers[C->c1].os == OS_WINGDOWS ||
			net.computers[C->c2].os == OS_WINGDOWS)
		{
			int dir, xdist, ydist,c;
			float sx, sy;
			dir = (net.computers[C->c2].os == OS_WINGDOWS);
			if (dir)
				{xdist=C->x1-C->x; ydist=C->y1-C->y;}
			else
				{xdist=C->x2-C->x; ydist=C->y2-C->y;}
			sx = xdist >= 0 ? 1.0 : -1.0;
			sy = ydist >= 0 ? 1.0 : -1.0;
			xdist = abs(xdist);
			ydist = abs(ydist);
			if (xdist==0 && ydist==0) {
				if (dir==0) c=C->c2; else c=C->c1;
				if (!net.computers[c].busy) {
					if (net.computers[c].os == OS_OFF)
						net.off--;
					else
						net.base--;
					net.win++;
					net.computers[c].os=OS_WINGDOWS;
				}
				C->active=0;
			}
			else if (MAX (xdist, ydist) < spark_speed) {
				if (dir)
					{C->x = C->x1; C->y = C->y1;}
				else
					{C->x = C->x2; C->y = C->y2;}
			}
			else {
				C->fx+=(xdist*spark_speed*sx)/(xdist+ydist);
				C->fy+=(ydist*spark_speed*sy)/(xdist+ydist);
				C->x = (int)C->fx;
				C->y = (int)C->fy;
			}
			C->index = 1 - C->index;
		}
	}
	else {
		if ((net.computers[C->c1].os == OS_WINGDOWS) ==
			(net.computers[C->c2].os == OS_WINGDOWS))
				C->delay = spark_delay(game.level);
		else if (net.computers[C->c1].os == OS_WINGDOWS ||
			net.computers[C->c2].os == OS_WINGDOWS)
		{
			if (C->delay>0) C->delay--;
			else {
				C->active = 1;
				if (net.computers[C->c1].os == OS_WINGDOWS)
					{C->fx=C->x=C->x1; C->fy=C->y=C->y1;}
				else
					{C->fx=C->x=C->x2; C->fy=C->y=C->y2;}
			}
		}
	}
}

int cable_onspark (Cable *C,int locx, int locy) {
	if (!C->active) return 0;
	return (abs(locx-C->x) < spark.width
		&& abs(locy-C->y) < spark.height);
}

void cable_draw(Cable *C) {
	int rx = C->x - spark.width/2;
	int ry = C->y - spark.height/2;
	ui_draw_line(C->x1,C->y1,C->x2,C->y2);
	if (C->active)
		ui_draw(&spark.pictures[C->index], rx, ry);
}
