#ifdef HAVE_CONFIG_H
#include "config.h"
#endif
#ifndef lint
static char rcsid[] ATTR_UNUSED = "$Id: game.c,v 1.1.1.1 1999/09/22 05:07:21 short Exp $";
#endif

#include <getopt.h>
#include <stdlib.h>
#include <assert.h>
#include <time.h>
#include <locale.h>
#include <errno.h>

#ifndef HAVE_GNOME
#ifdef HAVE_GETOPT_LONG
#include <getopt.h>
#else
#include "getopt.h"
#endif
#endif

#include "game.h"
#include "horde.h"
#include "network.h"
#include "library.h"
#include "bucket.h"
#include "spark.h"
#include "scorelist.h"
#include "ui.h"
#include "picture.h"
#include "data.h"
#ifdef HAVE_LIBGGN
#include "billggn.h"
#endif

Game game;

int game_RAND(int lb, int ub) {
	return (
#ifdef HAVE_LIBGGN
	(ggn_active()?ggn_Grand():rand())
#else
	rand()
#endif
	%(ub-lb+1) + lb);
}

int game_INTERSECT(int x1, int y1, int w1, int h1, int x2, int y2, int w2, int h2)
{
	return (((x2-x1<=w1 && x2-x1>=0) || (x1-x2<=w2 && x1-x2>=0))
		&& ((y2-y1<=h1 && y2-y1>=0) || (y1-y2<=h2 && y1-y2>=0)));
}

static struct deferrer {
	struct deferrer *next;
	void (*func)(int,int,char *); int x,y;
	char *text;
	} *defers,**deferstail=&defers;
static char defermode;

static void defer(void (*func)(int,int,char *),int x,int y,char *text)
{
struct deferrer *d;
	if (!(d=malloc(sizeof(**deferstail)))) return;
	d->next=NULL;
	d->func=func; d->x=x; d->y=y; d->text=text;
	*deferstail=d;
	deferstail=&d->next;
}

static void deferclear(void)
{
	assert(!defermode);
	defermode=2;
	while (defers) {
struct deferrer *d=defers;
		(*d->func)(d->x,d->y,d->text);
		if (!(defers=d->next)) deferstail=&defers;
		free(d);
		}
	defermode=0;
}

void game_setup_level (char restart,unsigned int lev) {
#ifdef HAVE_LIBGGN
	if (ggn_serving()) {
int newseed=rand();
		ggn_Gsrand(newseed);
		ggn_netprintf("%c%d,%d,\n",(restart?'S':'s'),newseed,lev);
		}
#endif
	if (restart) {
		game.state = game_PLAYING;
		game.score = 0;
		defermode=0;
		ui_restart_timer();
		ui_set_pausebutton(1);
		}
	game.level = lev;
	bill_setup();
	game.grabbed = game_EMPTY;
	ui_set_cursor(game_DEFAULTC);
	net_setup();
	game.iteration = game.efficiency = 0;
}

void game_start(unsigned int lev) {
	game_setup_level(1,lev);
}

void game_quit(void) {
	scores_write();
	exit(0);
}

void game_update_info(void) {
	static char str[80];
	sprintf (str, _("Bill:%d/%d  System:%d/%d/%d  Level:%d  Score:%d"),
		bill.on_screen, bill.off_screen, net.base, net.off,
		net.win, game.level, game.score);
	ui_set_statusbar(str);
	game.efficiency += ((100*net.base-10*net.win)/net.units);
}
  
void game_update_score (int action) {
	switch (action){
		case game_ENDLEVEL: game.score+=(game.level*game.efficiency/game.iteration); break;
		default: game.score+=(action*action*game_BILLPOINTS);
	}
}

void game_warp_to_level (int lev) {
	if (game.state==game_PLAYING) {
		if (lev <= game.level) return;
		game_setup_level(0,lev);
	}
	else {
		if (lev<=0) return;
		game_start(lev);
	}
}

static void game_draw_bang(int x,int y,char *text)
{ 
	ui_draw(&game.bang,x-game.bang.width/2,y-game.bang.height/2);
	if (text) ui_draw_string(text,x,y+game.bang.height/2,ui.yellowgc);
}
	
void game_button_press(int x, int y,char *text) {
	int i, counter=0, flag=0;
	if (game.state != game_PLAYING) return;
#ifdef HAVE_LIBGGN
	if (defermode==1 && ggn_serving()) {
		d(("%d: deferring button_press(%d,%d), it=%d\n",ggn.gc->cur,x,y,game.iteration));
		defer(game_button_press,x,y,text);
		return; }
	dbg(if (ggn_active())
		printf("%d: game_button_press(%d,%d), defermode=%d, it=%d\n",ggn.gc->cur,x,y,defermode,game.iteration));
	ggn_netprintf("p%d,%d,",x,y);
#endif
	ui_set_cursor(game_DOWNC);
	if (bucket_clicked(x, y)) {
		ui_set_cursor(game_BUCKETC);
		game.grabbed = game_BUCKET;
		flag=2;
	}
	for (i=0; i < bill_MAX_BILLS && flag!=1; i++) {
		if (bill.list[i].state == monster_OFF
			|| bill.list[i].state == monster_DYING)
				continue;
		if (bill.list[i].state == monster_STRAY &&
			monster_clickedstray(&bill.list[i],x,y))
		{
			ui_set_cursor (bill.list[i].cargo);
			game.grabbed = i;
			flag = 1;
		}
		else if (bill.list[i].state != monster_STRAY &&
			monster_clicked(&bill.list[i],x,y))
		{
			if (bill.list[i].state == monster_AT) 
				net.computers[bill.list[i].target_c].busy=0;
			bill.list[i].index = -1;
			bill.list[i].cels = bill.dcels;
			bill.list[i].x_offset = -2;
			bill.list[i].y_offset = -15;
			bill.list[i].state = monster_DYING;
			bill.list[i].killer=text;
			counter++;
#ifdef HAVE_LIBGGN
			dbg(if (ggn_active())
				printf("%d: killed bill %d in iteration %d\n",ggn.gc->cur,i,game.iteration));
#endif
		}
	}
	if (!flag) defer(game_draw_bang,x,y,text);
	if (!text && counter) game_update_score(counter);
}

void game_button_release(int x, int y,char *text) {
	int i;
	if (game.state != game_PLAYING)
		return;
#ifdef HAVE_LIBGGN
	if (defermode==1 && ggn_serving()) { defer(game_button_release,x,y,text); return; }
	ggn_netprintf("r%d,%d,",x,y);
#endif
	ui_set_cursor (game_DEFAULTC);
	if (game.grabbed == game_EMPTY)
		return;
	if (game.grabbed == game_BUCKET) {
		game.grabbed = game_EMPTY;
		for (i=0; i<net.ncables; i++)
			if (cable_onspark(&net.cables[i],x,y)) {
				net.cables[i].active=0;
				net.cables[i].delay = spark_delay(game.level);
			}
		return;
	}
	for (i=0; i<net.units; i++)
		if (computer_oncomputer(&net.computers[i],x,y)
			&&
			computer_compatible(&net.computers[i],bill.list[game.grabbed].cargo)
			&&
			(net.computers[i].os == OS_WINGDOWS || 
			 net.computers[i].os == OS_OFF))
		{
			net.base++; 
			if (net.computers[i].os == OS_WINGDOWS)
				net.win--;
			else
				net.off--;
			net.computers[i].os = bill.list[game.grabbed].cargo;
			bill.list[game.grabbed].state = monster_OFF;
			game.grabbed = game_EMPTY;
			return;
		}
	game.grabbed = game_EMPTY;
}

void game_clear(void)
{
	ui_clear();
	ui_draw_centered(&game.logo);
	ui_kill_timer();
	ui_set_pausebutton(0);
	game.state = game_WAITING;
}

void game_update(char fromnet) {
#ifdef HAVE_LIBGGN
	dbg(if (ggn_active()) printf("%d: game_update(fromnet=%d),it=%d begin\n",ggn.gc->cur,fromnet,game.iteration));
#endif
	switch (game.state) {
	case game_PLAYING:
#ifdef HAVE_LIBGGN
		if (ggn_active()) {
			if (!fromnet) {
				ggn_netprintf("\n");
				d(("%d: putting end-of-turn marker\n",ggn.gc->cur));
				defermode=1;
				}
			d(("%d: remaining=%d\n",ggn.gc->cur,ggn.remaining));
			if (ggn.remaining) {
				d(("%d: game_update(fromnet=%d),it=%d bailing as remaining=%d\n",
					ggn.gc->cur,fromnet,game.iteration,ggn.remaining));
				return;
				}
			if (!defermode) {
				d(("%d: game_update(fromnet=%d),it=%d bailing as !defermode\n",
					ggn.gc->cur,fromnet,game.iteration));
				return;
				}
			defermode=0;
			}
#endif
		ui_restart_timer();
		ui_clear();
		bucket_draw();
		net_update();
		net_draw();
		bill_update();
		bill_draw();
		game_update_info();
		if (!(bill.on_screen+bill.off_screen)) {
			game_update_score(game_ENDLEVEL);
			game.state = game_BETWEEN;
		}
		if ((net.base+net.off)<=1) game.state = game_END;
		game.iteration++;
#ifdef HAVE_LIBGGN
		ggn_resetmove();
		dbg(if (ggn_active()) printf("%d: game_update(fromnet=%d),it++=%d before defers\n",ggn.gc->cur,fromnet,game.iteration));
		{ char me=ggn.me; ggn.me=0;
#endif
			deferclear();
#ifdef HAVE_LIBGGN
			ggn.me=me;
		}
		dbg(if (ggn_active()) printf("%d: game_update(fromnet=%d),it++=%d after  defers\n",ggn.gc->cur,fromnet,game.iteration));
#endif
		break;
	case game_END:
		ui_clear();
		net_toasters();
		net_draw();
		ui_refresh();
		ui_popup_dialog(game_ENDGAME);
		if (game.score > scores.score[9]) ui_popup_dialog(game_ENTERNAME);
		scores_update();
		ui_popup_dialog(game_HIGHSCORE);
		game_clear();
		break;
	case game_BETWEEN:
#ifdef HAVE_LIBGGN
		dbg(if (ggn_active()) printf("%d: BETWEEN begin\n",ggn.gc->cur));
#endif
		ui_update_scorebox(game.level,game.score);
		game.state=game_PLAYING;
#ifdef HAVE_LIBGGN
		{ char me=ggn.me; ggn.me=1;
#endif
			game_setup_level(0,++game.level);
#ifdef HAVE_LIBGGN
			ggn.me=me;
		}
		ggn_resetmove();
#endif
		ui_popup_dialog(game_SCORE);
		ui_restart_timer();
#ifdef HAVE_LIBGGN
		dbg(if (ggn_active()) printf("%d: BETWEEN finish\n",ggn.gc->cur));
#endif
		break;
	}
	ui_refresh();
}

/* __NORETURN */ void game_usage(void)
{
	fprintf(stderr,_("\
%s\n\
This command implements xBill game:\n\
\n\
Usage: xbill [-l <level>] [-h|--help] [-V|--version]\n\
\n\
  -l, --level\t\tStart at level <level>\n\
  -h, --help\t\tPrint a summary of the options and exit\n\
  -V, --version\t\tPrint the version number and exit\n\
"),game.version);
	exit(EXIT_FAILURE);
}

#ifndef HAVE_GNOME
const struct option longopts[]={
{"level"  ,1,0,'l'},
{"help"   ,0,0,'h'},
{"version",0,0,'V'}};

#else

struct poptOption game_popts[]=
{
	{ "level",'l',POPT_ARG_INT,&game.level,0,N_("Start at desired level"),N_("LEVEL") },
	{ NULL }
};

#endif

int main(int argc, char **argv) {
#ifndef HAVE_GNOME
	int optc;
	char *s;
#endif

	game.version="xBill " VERSION;

	/* Initialize the i18n stuff */
	bindtextdomain(PACKAGE,LOCALEDIR);
#ifndef HAVE_GNOME
	if (!setlocale(LC_ALL,""))
		fprintf(stderr,"%s: Locale not supported by C library!\n",argv[0]);
#endif
	textdomain(PACKAGE);

#ifndef HAVE_GNOME
	while ((optc=getopt_long(argc,argv,"l:hV",longopts,NULL))!=EOF) switch (optc) {
		case 'l':
			errno=EINVAL;
			game.level=strtol(optarg,&s,0);
			if (*s!='\0'||game.level<1||game.level>10000)
				{ perror(optarg); game_usage(); }
			break;
		case 'V':
			fputs(game.version,stderr);
			fputc('\n',stderr);
			exit(EXIT_FAILURE);
		default: /* also 'h' */
			game_usage();
			break;
		}
	if (optind<argc) {
		fprintf(stderr,_("Excessive %d arguments!\n"),argc-optind);
		game_usage();
		}
#endif

	setlinebuf(stdout);
	data_load();
	ui_initialize(&argc, &argv);
	srand(time(NULL));
	ui_clear();
	PICTURE_LOAD(game.logo,logo);
	ui_draw_centered(&game.logo);
	ui_refresh();
	ui_make_windows();
	scores_read();
	scores_update();

	PICTURE_LOAD(game.bang,bang);
	bill_load_pix();
	OS_load_pix();
	net_load_pix();
	bucket_load_pix();
	spark_load_pix();

	mcursor_finish();
	data_finish();

#ifdef HAVE_LIBGGN
	ggn_setup();
#endif

	game.state = game_WAITING;
	if (game.level) game_start(game.level);
	else ui_set_pausebutton(0);
	ui_set_statusbar(_("Initialized and ready"));
	ui_MainLoop();
	exit(0);
}
