#ifdef HAVE_CONFIG_H
#include "config.h"
#endif
#ifndef lint
static char rcsid[] ATTR_UNUSED = "$Id: monster.c,v 1.1.1.1 1999/09/22 05:07:21 short Exp $";
#endif

#include <glib.h>

#include "monster.h"
#include "game.h"
#include "network.h"
#include "ui.h"
#include "horde.h"
#include "library.h"

void monster_get_border(Monster *M) {
	int i=game_RAND(0,3);
	if (i%2==0) M->target_x=game_RAND(0, game_scrwidth-bill.width);
	else M->target_y=game_RAND(0, game_scrheight-bill.height);
	switch (i) {
		case 0: M->target_y=-bill.height-16; break;
		case 1: M->target_x=game_scrwidth+1; break;
		case 2: M->target_y=game_scrwidth+1; break;
		case 3: M->target_x=-bill.width-2; break;
	}
}

/*  Adds a bill to the in state*/
void monster_enter(Monster *M){
	M->state = monster_IN;
	monster_get_border(M);
	M->x = M->target_x; M->y=M->target_y;
	M->index = 0;
	M->cels = bill.lcels;
	M->cargo = OS_WINGDOWS;
	M->x_offset = -2;
	M->y_offset = -15;
	M->target_c = game_RAND(0, net.units-1);
	M->target_x = net.computers[M->target_c].x+net.width-monster_XOFFSET;
	M->target_y = net.computers[M->target_c].y+monster_YOFFSET;
	M->killer=NULL;
	bill.on_screen++; bill.off_screen--;
}

/*  Moves bill toward his target - returns whether or not he moved */
int monster_move(Monster *M,int mode) {
	int xdist = M->target_x - M->x;
	int ydist = M->target_y - M->y;
	int step = monster_step_size(game.level);
	int dx, dy;
	int signx = xdist >= 0 ? 1 : -1;
	int signy = ydist >= 0 ? 1 : -1;
	xdist = abs(xdist);
	ydist = abs(ydist);
	if (!xdist && !ydist) return 0;
	else if (xdist<step && ydist<step) {
		M->x = M->target_x;
		M->y = M->target_y;
	}
	else {
		dx = (xdist*step*signx)/(xdist+ydist);
		dy = (ydist*step*signy)/(xdist+ydist);
		switch(mode) {
			case monster_SLOW: break;
			case monster_FAST: dx = 5*dx/4; dy = 5*dy/4; break;
		}
		M->x+=dx;
		M->y+=dy;
		if (dx<0)
			M->cels = bill.lcels;
		else if (dx>0)
			M->cels = bill.rcels;
	}
	return 1;
}

void monster_draw(Monster *M) {
	switch (M->state) {
		case monster_IN:
		case monster_OUT:
		case monster_DYING: monster_draw_std(M);   break;
		case monster_AT:    monster_draw_at(M);    break;
		case monster_STRAY: monster_draw_stray(M); break;
		default: break;
	}
}

/*  Update Bill's position */	
void monster_update_in(Monster *M) {
	int moved = monster_move(M,monster_SLOW);
	if (!moved && (net.computers[M->target_c].os != OS_WINGDOWS)
		&& !(net.computers[M->target_c].busy))
	{
		net.computers[M->target_c].busy=1;
		M->cels = bill.acels;
		M->index=0;
		M->state = monster_AT;
		return;
	}
	else if (!moved) {
		int i;
		do {
			i=game_RAND(0, net.units-1);
		} while (i == M->target_c);
		M->target_c = i;
		M->target_x = net.computers[M->target_c].x + net.width-monster_XOFFSET;
		M->target_y = net.computers[M->target_c].y + monster_YOFFSET;
	}
	M->index++;
	M->index%=bill_WCELS;
	M->y_offset+=(8*(M->index%2)-4);
}

/*  Update Bill standing at a computer  */
void monster_update_at(Monster *M) {
	int sys;
	if (M->index==0 && net.computers[M->target_c].os == OS_OFF) {
		M->index=6;
		sys = computer_find_stray(&net.computers[M->target_c]);
		if (sys<0) M->cargo = -1;
		else {
			M->cargo = bill.list[sys].cargo;
			bill.list[sys].state = monster_OFF;
		}
	}
	else M->index++;
	if (M->index == 13) {
		M->y_offset = -15;
		M->x_offset = -2;
		monster_get_border(M);
		M->index = 0;
		M->cels = bill.lcels;
		M->state = monster_OUT;
		net.computers[M->target_c].busy=0;
		return;
	}
	M->y_offset = bill.height - OS.height;
	switch (M->index) {
	case  1 : 
	case  2 : M->x -=  8; M->x_offset += 8; break;
	case  3 : M->x -= 10; M->x_offset +=10; break;
	case  4 : M->x +=  3; M->x_offset -= 3; break;
	case  5 : M->x +=  2; M->x_offset -= 2; break;
	case  6 :
		if (net.computers[M->target_c].os != OS_OFF) {
			net.base--; net.off++;
			M->cargo = net.computers[M->target_c].os;
		}
		else {
			M->x-=21; M->x_offset+=21;
		}
		net.computers[M->target_c].os = OS_OFF;
		M->y_offset = -15;
		M->x += 20;
		M->x_offset -=20;
		break;
	case  7 : M->sy = M->y_offset; M->sx = -2; break;
	case  8 : M->sy =         -15; M->sx = -2; break;
	case  9 : M->sy =          -7; M->sx = -7; M->x -=  8; M->x_offset += 8; break;	
	case 10 : M->sy =           0; M->sx = -7; M->x -= 15; M->x_offset +=15; break;
	case 11 : M->sy =           0; M->sx = -7;
		net.computers[M->target_c].os = OS_WINGDOWS;
		net.off--; net.win++;
		break;
	case 12 : M->x += 11; M->x_offset -=11;
	}
}

/* Updates Bill fleeing with his ill gotten gain */
void monster_update_out(Monster *M) {
	if (game_INTERSECT(M->x,M->y,bill.width,bill.height,0,0,game_scrwidth,game_scrheight))
	{
		monster_move(M,monster_FAST);
		M->index++;
		M->index%=bill_WCELS;
		M->y_offset+=(8*(M->index%2)-4); 
	}
	else {
		M->state = monster_OFF;
		bill.on_screen--; bill.off_screen++;
	}
}


/* Updates Bill who is dying */
void monster_update_dying(Monster *M) {
	if (M->index < bill_DCELS-1){
		M->y_offset += (M->index*monster_GRAVITY);
		M->index++;	
	}
	else {
		M->y+=M->y_offset;
		if (M->cargo<0 || M->cargo == OS_WINGDOWS) M->state = monster_OFF;
		else M->state = monster_STRAY;
		bill.on_screen--;
	}
}

void monster_update(Monster *M) {
	switch (M->state) {
		case monster_IN   : monster_update_in   (M); break;
		case monster_AT   : monster_update_at   (M); break;
		case monster_OUT  : monster_update_out  (M); break;
		case monster_DYING: monster_update_dying(M); break;
		default: break;
	}
}

int monster_clicked(Monster *M,int locx, int locy) {
	return (locx>M->x && locx<M->x+bill.width && locy>M->y && locy<M->y+bill.height);
}

int monster_clickedstray(Monster *M,int locx, int locy) {
	return (locx>M->x && locx<M->x+OS.width && locy>M->y && locy<M->y+OS.height);
}

int monster_step_size(unsigned int lev) {
	return MIN(14+lev, 18);
}

void monster_draw_std(Monster *M) {
	if (M->cargo>=0)
		ui_draw(&OS.os[M->cargo], M->x + M->x_offset, M->y + M->y_offset);
	ui_draw(&M->cels[M->index], M->x, M->y);
	if (M->state==monster_DYING && M->killer)
		ui_draw_string(M->killer,M->x+M->cels[M->index].width/2,
					M->y+M->cels[M->index].height,ui.redgc);
}

void monster_draw_at(Monster *M) {
	if (M->index>6 && M->index<12)
		ui_draw(&OS.os[0], M->x + M->sx, M->y + M->sy);
	if (M->cargo>=0)
		ui_draw(&OS.os[M->cargo], M->x + M->x_offset, M->y + M->y_offset);
	ui_draw(&M->cels[M->index], net.computers[M->target_c].x,
		net.computers[M->target_c].y);
}

void monster_draw_stray(Monster *M) {
	if (game.grabbed==-1 || M->x != bill.list[game.grabbed].x)
		ui_draw(&OS.os[M->cargo], M->x, M->y);
}
