#ifdef HAVE_CONFIG_H
#include "config.h"
#endif
#ifndef lint
static char rcsid[] ATTR_UNUSED = "$Id: ui.c,v 1.2 1999/10/10 08:48:17 short Exp $";
#endif

#include <gtk/gtk.h>
#include <gdk/gdkkeysyms.h>
#include <stdio.h>
#include <assert.h>
#include <string.h>
#include <stdlib.h>

#ifdef HAVE_GNOME
#include <gnome.h>
#endif

#include "gtk10.h"
#include "ui.h"
#include "game.h"
#include "scorelist.h"
#include "bucket.h"
#include "library.h"
#include "picture.h"
#include "mcursor.h"
#ifdef HAVE_LIBGGN
#include "billggn.h"
#endif

/***********************************
* Strings originally from strings.h *
 ***********************************/

static char *storystr=N_("\
The Story:\n\
\n\
Yet again, the fate of the world rests\n\
in your hands! An evil computer hacker,\n\
known only by his handle 'Bill', has\n\
created the ultimate computer virus. A\n\
virus so powerful that it has the power\n\
to transmute an ordinary computer into\n\
a toaster oven. (oooh!) 'Bill' has\n\
cloned himself into a billion-jillion\n\
micro-Bills. Their sole purpose is to\n\
deliver the nefarious virus, which has\n\
been cleverly diguised as a popular\n\
operating system.\n\
\n\
As System Administrator/Exterminator,\n\
your job is to keep Bill from succeeding\n\
at his task.");

static char *rulesstr=N_("\
The Rules:\n\
\n\
xBill has been painstakingly designed and\n\
researched in order to make it as easy to use\n\
for the whole family as it is for little Sally.\n\
Years - nay - days of beta testing and \n\
consulting with the cheapest of human interface\n\
designers have resulted in a game that is easy\n\
to use, yet nothing at all like a Macintosh.\n\
\n\
I.   Whack the Bills (click)\n\
II.  Restart the computer (click)\n\
III. Pick up stolen OSes & return(drag)\n\
     them to their respective computers\n\
IV.  Drag the bucket to extinguish sparks\n\
V.   Scoring is based on total uptime,\n\
     with bonuses for killing Bills.\n\
\n\
As for the rest, you can probably figure\n\
it out. We did, so it can't be too hard.");

static char *endgamestr=N_("\
Module xBill has caused a segmentation fault\n\
at memory address 097E:F1A0. Core dumped.\n\
\n\
We apologize for the inconvenience.");

/***********************************/

UI ui;

static void newgamefn    (void)
{ game_start(1); }

#ifdef HAVE_LIBGGN
static void ggnlaunchfn  (void)
{ ui_pause_game(); ggn_menu(); }
static void ggndisconnfn (void)
{ ggn_disconn(); }
#endif

#define GENSHOW(name) static void s##name(void) { gtk_widget_show(GTK_WIDGET(ui.name)); }
GENSHOW(pausebox)
GENSHOW(levelwin)
GENSHOW(highscorebox)
GENSHOW(storybox)
GENSHOW(rulesbox)
GENSHOW(aboutbox)
#undef GENSHOW

static void quitfn       (void)
{ game_quit(); }

#ifndef HAVE_GNOME
static GtkItemFactoryEntry menu[]=
{
	{N_("/_Game")                 ,NULL        ,(GtkItemFactoryCallback)NULL         ,0,"<Branch>"   },
	{N_("/Game/tearoff1")         ,NULL        ,(GtkItemFactoryCallback)NULL         ,0,"<Tearoff>"  },
	{N_("/Game/_New game")        ,"<control>N",(GtkItemFactoryCallback)newgamefn    ,0,NULL         },
#ifdef HAVE_LIBGGN
	{N_("/Game/Net_work play...") ,"<control>W",(GtkItemFactoryCallback)ggnlaunchfn  ,0,NULL         },
	{N_("/Game/_Disconnect...")   ,"<control>D",(GtkItemFactoryCallback)ggndisconnfn ,0,NULL         },
#endif
	{N_("/Game/_Pause game")      ,"<control>P",(GtkItemFactoryCallback)spausebox    ,0,NULL         },
	{N_("/Game/Warp to _level..."),"<control>L",(GtkItemFactoryCallback)slevelwin    ,0,NULL         },
	{N_("/Game/View _high scores"),"<control>H",(GtkItemFactoryCallback)shighscorebox,0,NULL         },
	{N_("/Game/sep1")             ,NULL        ,(GtkItemFactoryCallback)NULL         ,0,"<Separator>"},
	{N_("/Game/_Quit game")       ,"<control>Q",(GtkItemFactoryCallback)quitfn       ,0,NULL         },
	{N_("/_Help")                 ,NULL        ,(GtkItemFactoryCallback)NULL         ,0,"<Branch>"   },
	{N_("/Help/_Story of xBill")  ,"<control>S",(GtkItemFactoryCallback)sstorybox    ,0,NULL         },
	{N_("/Help/_Rules")           ,"<control>R",(GtkItemFactoryCallback)srulesbox    ,0,NULL         },
	{N_("/Help/_About xBill")     ,"<control>A",(GtkItemFactoryCallback)saboutbox    ,0,NULL         },
};
#else /* HAVE_GNOME */

#define ACCEL_UI(label,accel,tooltip,callback) \
	{ GNOME_APP_UI_ITEM, label, tooltip, callback, NULL, NULL, \
	GNOME_APP_PIXMAP_NONE, NULL, accel, GDK_CONTROL_MASK, NULL }

static GnomeUIInfo menu_game[]=
{
	GNOMEUIINFO_MENU_NEW_GAME_ITEM((void *)newgamefn,0),
#ifdef HAVE_LIBGGN
	ACCEL_UI(N_("Net_work play...") ,'W',N_("Pops up network menu")                   ,(void *)ggnlaunchfn),
	ACCEL_UI(N_("_Disconnect...")   ,'D',N_("Terminate currently active network play"),(void *)ggndisconnfn),
#define PAUSE_MENU_NUM 3
#else
#define PAUSE_MENU_NUM 1
#endif

	GNOMEUIINFO_MENU_PAUSE_GAME_ITEM((void *)spausebox,0),
	ACCEL_UI(N_("Warp to _level..."),'L',N_("Choose arbitrary level to play")          ,(void *)slevelwin),
	GNOMEUIINFO_MENU_SCORES_ITEM((void *)shighscorebox,0),
	GNOMEUIINFO_SEPARATOR,
	GNOMEUIINFO_MENU_EXIT_ITEM((void *)quitfn,0),
	GNOMEUIINFO_END
};

static GnomeUIInfo menu_help[]=
{
	ACCEL_UI(N_("_Story of xBill"),'S',N_("Motivation of the game playing"),(void *)sstorybox),
	ACCEL_UI(N_("_Rules")         ,'R',N_("Written rules of the playing")  ,(void *)srulesbox),
	GNOMEUIINFO_MENU_ABOUT_ITEM((void *)saboutbox,0),
	GNOMEUIINFO_END
};

static GnomeUIInfo menu[]=
{
	GNOMEUIINFO_MENU_GAME_TREE(menu_game),
	GNOMEUIINFO_MENU_HELP_TREE(menu_help),
	GNOMEUIINFO_END
};

#endif /* HAVE_GNOME */

/**************************/
/* Timer control routines */
/**************************/

static gint timer_eh(gpointer data)
{
	if (ui.istimer) {
		ui_kill_timer();
/*		if (ggn_active()) printf("%d: timer_eh()\n",ggn.gc->cur);*/
		game_update(0);
		}
	return(FALSE);
}

void ui_restart_timer(void) {
/*	if (ggn_active()) printf("%d: restart_timer()\n",ggn.gc->cur);*/
	ui_kill_timer();
	ui.timer = gtk_timeout_add(250,timer_eh,NULL); /* 250 ms */
	ui.istimer=1;
}

void ui_kill_timer(void) {
	if (ui.istimer) {
/*		if (ggn_active()) printf("%d: kill_timer() - killing it, eek\n",ggn.gc->cur);*/
		gtk_timeout_remove(ui.timer);
		ui.istimer=0;
	}
}

void ui_pause_game(void) {
	if (ui.istimer) ui.playing = 1;
	ui_kill_timer();
}

void ui_resume_game(void) {
	if (ui.playing) ui_restart_timer();
	ui.playing = 0;
}

/*******************/
/* Window routines */
/*******************/

static GdkGC *newgc(GdkWindow *win,gushort red,gushort green,gushort blue)
{
GdkGC *r;
GdkColor col;
	col.red  =red  ;
	col.green=green;
	col.blue =blue ;
	gdk_color_alloc(gdk_window_get_colormap(win),&col);
	r=gdk_gc_new(ui.offscreen);
	gdk_gc_set_foreground(r,&col);
	assert(r);
	return(r);
}

/* Create a new backing pixmap of the appropriate size */
static void configure_event(GtkWidget *widget,GdkEventConfigure *event)
{
static int aw=-1,ah=-1;
	if (aw==widget->allocation.width && ah==widget->allocation.height) return;
	    aw= widget->allocation.width;   ah= widget->allocation.height;
	if (ui.offscreen) gdk_pixmap_unref(ui.offscreen);
	ui.offscreen=gdk_pixmap_new(widget->window,
		widget->allocation.width,
		widget->allocation.height,
		-1);
	if (!ui.linegc) {
		ui.linegc=gdk_gc_new(ui.offscreen);
		gdk_gc_set_line_attributes(ui.linegc,3,GDK_LINE_SOLID,GDK_CAP_ROUND,GDK_JOIN_MITER);
		gdk_gc_set_foreground(ui.linegc,&ui.mainwinstyle->black);
		assert(ui.linegc);
		}
	if (!ui. whitegc) ui. whitegc=newgc(widget->window,0xFFFF,0xFFFF,0xFFFF);
	if (!ui.yellowgc) ui.yellowgc=newgc(widget->window,0xFFFF,0xFFFF,0x0000);
	if (!ui.   redgc) ui.   redgc=newgc(widget->window,0xFFFF,0x0000,0x0000);
	ui_clear();
}

/* Redraw the screen from the backing pixmap */
static void expose_event(GtkWidget *widget,GdkEventExpose *event)
{
	gdk_draw_pixmap(widget->window,
		widget->style->black_gc,
		ui.offscreen,
		event->area.x, event->area.y,
		event->area.x, event->area.y,
		event->area.width, event->area.height);
}

static void button_press_event(GtkWidget *widget,GdkEventButton *event)
{
	/* Ignore double and triple click: */
	if (event->type!=GDK_BUTTON_PRESS) return;
	game_button_press(event->x,event->y,NULL);
}

static void button_release_event(GtkWidget *widget,GdkEventButton *event)
{ game_button_release(event->x,event->y,NULL); }

static void enter_notify_event(GtkWidget *widget,GdkEventCrossing *event)
{ ui_resume_game(); }

static void leave_notify_event(GtkWidget *widget,GdkEventCrossing *event)
{
#ifdef HAVE_LIBGGN
	if (ggn_active()) return;
#endif
	ui_pause_game();
}

static gint entry_keypress(GtkWidget *widget,GdkEventKey *key,GtkSignalFunc func)
{ if (key->keyval==GDK_Return) (*func)(); return(1); }

static GtkWidget *newentry(const char *label)
{
GtkWidget *hbox,*r;

	hbox=gtk_hbox_new(FALSE,10);
		{ GtkWidget *lab;
		lab=gtk_label_new(label);
		gtk_box_pack_start(GTK_BOX(hbox),lab,FALSE,FALSE,0);
		gtk_widget_show(lab);
		}

		{ GtkWidget *ent;
			ent=gtk_entry_new();
			gtk_box_pack_start_defaults(GTK_BOX(hbox),ent);
			gtk_widget_show(ent);
			r=ent;
		}
	gtk_widget_show(hbox);
	return(r);
}

static int levelvalue(void)
{ return atoi(gtk_entry_get_text(GTK_ENTRY(ui.levelentry))); }

static void warpnow(void)
{
/* Don't use ->adjustment value as otherwise MAX_LEVEL couldn't be exceeded! */
int lev=levelvalue();
	if (lev<=0 || lev>=INT_MAX);
	game_warp_to_level(lev);
	gtk_widget_hide(ui.levelwin);
}

static gint levelentry_keyrelease(GtkWidget *widget,GdkEventKey *key,GtkSignalFunc func)
{
int lev=levelvalue();
	if (lev>=1 && lev<=MAX_LEVEL)
		gtk_adjustment_set_value(GTK_SPIN_BUTTON(ui.levelentry)->adjustment,(gfloat)lev);
	return(1);
}

GtkWidget *vhbutton(const char *label,GtkWidget *vbox)
{
GtkWidget *hbox,*butt;
	hbox=gtk_hbutton_box_new();
	butt=gtk_button_new_with_label(label?label:_("Okay"));
	gtk_container_set_border_width(GTK_CONTAINER(butt),2);
	gtk_box_pack_start_defaults(GTK_BOX(hbox),butt);
	gtk_widget_show(butt);
	if (vbox) gtk_box_pack_start_defaults(GTK_BOX(vbox),hbox);
	gtk_widget_show(hbox);
	return(butt);
}

static GtkWidget *newentrybox(const char *label,const char *oktext,GtkWidget *entry,GtkWidget **vboxp,GtkSignalFunc func)
{
GtkWidget *win,*fr,*vbox;
	win=gtk_dialog_new();
	gtk_window_set_position(GTK_WINDOW(win),GTK_WIN_POS_MOUSE);
	gtk_window_set_title(GTK_WINDOW(win),label);
	gtk_container_set_border_width(GTK_CONTAINER(win),10);
	fr=gtk_frame_new(NULL);
	gtk_box_pack_start_defaults(GTK_BOX(GTK_DIALOG(win)->vbox),fr);
	gtk_widget_show(fr);
	vbox=gtk_vbox_new(FALSE,5);
	gtk_container_add(GTK_CONTAINER(fr),vbox);
	gtk_container_set_border_width(GTK_CONTAINER(vbox),10);
	gtk_widget_show(vbox);
	*vboxp=vbox;
	gtk_signal_connect(GTK_OBJECT(entry),"key_press_event",GTK_SIGNAL_FUNC(entry_keypress),(gpointer)func);
	gtk_signal_connect_object(GTK_OBJECT(win),"show",GTK_SIGNAL_FUNC(gtk_widget_grab_focus),GTK_OBJECT(entry));
		{ GtkWidget *bok;
		bok=gtk_button_new_with_label(oktext);
		gtk_signal_connect(GTK_OBJECT(bok),"clicked",func,entry);
		gtk_box_pack_start_defaults(GTK_BOX(GTK_DIALOG(win)->action_area),bok);
		gtk_widget_show(bok);
		}
		{ GtkWidget *bcancel;
		bcancel=gtk_button_new_with_label(_("Cancel"));
		gtk_signal_connect_object(GTK_OBJECT(bcancel),"clicked",GTK_SIGNAL_FUNC(gtk_widget_hide),GTK_OBJECT(win));
		gtk_box_pack_start_defaults(GTK_BOX(GTK_DIALOG(win)->action_area),bcancel);
		gtk_widget_show(bcancel);
		}
	gtk_widget_show(entry);
	return(win);
}

void ui_initialize(int *argc, char ***argv) {
	GtkWidget *mainv,*fixed,*hsep,*box;
#ifndef HAVE_GNOME
	GtkWidget *menuwidget;
#ifdef HAVE_GTK_ITEM_FACTORY_CREATE_ITEMS
	GtkAccelGroup *accel;
#endif /* HAVE_GTK_ITEM_FACTORY_CREATE_ITEMS */

	gtk_init(argc,argv);
#else /* HAVE_GNOME */
	gnome_init_with_popt_table("xbill",VERSION,*argc,*argv,game_popts,0,NULL);
#endif /* HAVE_GNOME */

#ifdef HAVE_GNOME
	ui.mainwin=gnome_app_new("xbill",game.version);
#else
	ui.mainwin=gtk_window_new(GTK_WINDOW_TOPLEVEL);
#endif
	gtk_window_set_policy(GTK_WINDOW(ui.mainwin),FALSE,FALSE,TRUE);
	ui.mainwinstyle=gtk_widget_get_style(ui.mainwin);
	gtk_signal_connect(GTK_OBJECT(ui.mainwin),"destroy"     ,GTK_SIGNAL_FUNC(gtk_main_quit),NULL);
	gtk_signal_connect(GTK_OBJECT(ui.mainwin),"delete_event",GTK_SIGNAL_FUNC(gtk_main_quit),NULL);

	{ GtkWidget *hbox,*hscale,*lab; GtkObject *adj;

	hbox=gtk_hbox_new(FALSE,30);
	lab=gtk_label_new(_("Level:"));
	gtk_widget_show(lab);
	gtk_box_pack_start_defaults(GTK_BOX(hbox),lab);
	adj=gtk_adjustment_new(1.0,1.0,(gfloat)(MAX_LEVEL+MAX_LEVEL_STEP),1.0,(gfloat)MAX_LEVEL_STEP,(gfloat)MAX_LEVEL_STEP);
	ui.levelentry=gtk_spin_button_new(GTK_ADJUSTMENT(adj),1.0,0);
	gtk_signal_connect(GTK_OBJECT(ui.levelentry),"key_release_event",GTK_SIGNAL_FUNC(levelentry_keyrelease),NULL);
	gtk_widget_set_usize(ui.levelentry,200,0);
	gtk_spin_button_set_numeric(GTK_SPIN_BUTTON(ui.levelentry),TRUE);
	hscale=gtk_hscale_new(GTK_ADJUSTMENT(adj));
	gtk_scale_set_digits(GTK_SCALE(hscale),0);
	gtk_widget_show(hscale);
	gtk_box_pack_start_defaults(GTK_BOX(hbox),ui.levelentry);
	gtk_widget_show(hbox);

	ui.levelwin=newentrybox(_("Level Warp"),_("Warp!"),ui.levelentry,&box,GTK_SIGNAL_FUNC(warpnow));
	gtk_box_pack_start_defaults(GTK_BOX(box),hbox);
	gtk_box_pack_start_defaults(GTK_BOX(box),hscale);
	}

	mainv=gtk_vbox_new(FALSE,0);
#ifdef HAVE_GNOME
	gnome_app_set_contents(GNOME_APP(ui.mainwin),mainv);
#else
	gtk_container_add(GTK_CONTAINER(ui.mainwin),mainv);
#endif
	gtk_widget_show(mainv);
	gtk_window_set_title(GTK_WINDOW(ui.mainwin),game.version);

#ifdef HAVE_GNOME
	gnome_app_create_menus(GNOME_APP(ui.mainwin),menu);
#else /* GAVE_GNOME */
	{ int i;
		for (i=0;i<NSIZE(menu);i++) menu[i].path=_(menu[i].path);
	}
#ifdef HAVE_GTK_ITEM_FACTORY_CREATE_ITEMS
	accel=gtk_accel_group_new();
	ui.mainwinfactory=gtk_item_factory_new(GTK_TYPE_MENU_BAR,"<Main>",accel);
	gtk_item_factory_create_items(ui.mainwinfactory,NSIZE(menu),menu,NULL);
	gtk_accel_group_attach(accel,GTK_OBJECT(ui.mainwin));
	menuwidget=gtk_item_factory_get_widget(ui.mainwinfactory,"<Main>");
#else /* !HAVE_GNOME && HAVE_GTK_ITEM_FACTORY_CREATE_ITEMS */
	menuwidget=gtk10_item_factory_create_items(NSIZE(menu),menu,"<Main>",ui.mainwin);
#endif /* !HAVE_GNOME && HAVE_GTK_ITEM_FACTORY_CREATE_ITEMS */
	gtk_box_pack_start_defaults(GTK_BOX(mainv),menuwidget);
	gtk_widget_show(menuwidget);
#endif /* HAVE_GNOME */

	ui.drawarea=gtk_drawing_area_new();
	gtk_drawing_area_size(GTK_DRAWING_AREA(ui.drawarea),game_scrwidth,game_scrheight);

#define ADDHANDLER(name) gtk_signal_connect(GTK_OBJECT(ui.drawarea),#name "_event",(GtkSignalFunc)name##_event,NULL)

	ADDHANDLER(expose);
	ADDHANDLER(configure);
	ADDHANDLER(button_press);
	ADDHANDLER(button_release);
	ADDHANDLER(enter_notify);
	ADDHANDLER(leave_notify);

#undef ADDHANDLER
	gtk_widget_set_events (ui.drawarea,0
		| GDK_EXPOSURE_MASK
		| GDK_BUTTON_PRESS_MASK
		| GDK_BUTTON_RELEASE_MASK
		| GDK_ENTER_NOTIFY_MASK
		| GDK_LEAVE_NOTIFY_MASK
		);
        gtk_box_pack_start_defaults(GTK_BOX(mainv),ui.drawarea);
	gtk_widget_show(ui.drawarea);

	hsep=gtk_hseparator_new();
	gtk_box_pack_start_defaults(GTK_BOX(mainv),hsep);
	gtk_widget_show(hsep);

	fixed=gtk_fixed_new();
	gtk_box_pack_start_defaults(GTK_BOX(mainv),fixed);
	gtk_container_set_border_width(GTK_CONTAINER(fixed),2);
	ui.statusbar=gtk_label_new(_("Setting up..."));
	gtk_fixed_put(GTK_FIXED(fixed),ui.statusbar,0,0);
	gtk_widget_show(ui.statusbar);
	gtk_widget_show(fixed);

	gtk_widget_show(ui.mainwin);

	MCURSOR_LOAD_SEP(ui.defaultcursor,hand_up);
	mcursor_set(&ui.defaultcursor);
	MCURSOR_LOAD_SEP(ui.downcursor,hand_down);
}

void ui_set_statusbar(const char *str)
{
static char *orig=NULL;
	if (orig && !strcmp(orig,str)) return;
	free(orig);
	orig=strdup(str);
	gtk_label_set_text(GTK_LABEL(ui.statusbar),str);
}

static GtkWidget *CreatePixmapBox(const char *name,Picture *picp,const char *text)
{
GtkWidget *win,*logopix,*butt;

	win=gtk_dialog_new();
	gtk_window_set_position(GTK_WINDOW(win),GTK_WIN_POS_MOUSE);
	gtk_window_set_title(GTK_WINDOW(win),name);
	gtk_container_set_border_width(GTK_CONTAINER(win),10);
	if (((int)picp)!=-1) {
		logopix=picture_gtk_pixmap(&game.logo);
		gtk_box_pack_start_defaults(GTK_BOX(GTK_DIALOG(win)->vbox),logopix);
		gtk_widget_show(logopix);
		if (picp) { GtkWidget *picpix;
			picpix=picture_gtk_pixmap(picp);
			gtk_box_pack_start_defaults(GTK_BOX(GTK_DIALOG(win)->vbox),picpix);
			gtk_widget_show(picpix);
			}
		}
	if (text) { GtkWidget *txt,*fr,*vbox;
		fr=gtk_frame_new(NULL);
		vbox=gtk_vbox_new(FALSE,5);
		gtk_container_set_border_width(GTK_CONTAINER(vbox),5);
		txt=gtk_label_new(text);
		gtk_widget_show(txt);
		gtk_box_pack_start(GTK_BOX(vbox),txt,FALSE,TRUE,5);
		gtk_widget_show(vbox);
		gtk_container_add(GTK_CONTAINER(fr),vbox);
		gtk_widget_show(fr);
		gtk_box_pack_start(GTK_BOX(GTK_DIALOG(win)->vbox),fr,FALSE,TRUE,10);
		ui.lastCPBtext=txt;
		}

	butt=gtk_button_new_with_label(_("Okay"));
	gtk_container_set_border_width(GTK_CONTAINER(butt),2);
	gtk_box_pack_start_defaults(GTK_BOX(GTK_DIALOG(win)->action_area),butt);
	gtk_signal_connect_object(GTK_OBJECT(butt),"clicked",GTK_SIGNAL_FUNC(gtk_widget_hide),GTK_OBJECT(win));
	gtk_widget_show(butt);
	gtk_window_set_focus(GTK_WINDOW(win),butt);
	return(win);
}

static void enternamefunc(void)
{
char *s=gtk_entry_get_text(GTK_ENTRY(ui.enternameentry)),buf[21];
	if (*s) {
		strncpy(buf,s,sizeof(buf));
		buf[sizeof(buf)-1]='\0';
		scores_recalc(buf);
		}
	else scores_recalc(_("Anonymous"));
	gtk_widget_hide(ui.enternamebox);
}

static void modalquit(GtkWidget *win)
{
	gtk_signal_connect(GTK_OBJECT(win),"hide",GTK_SIGNAL_FUNC(gtk_main_quit),NULL);
	gtk_window_set_modal(GTK_WINDOW(win),TRUE);
}

static Picture aboutpic;

void ui_make_windows() {
GtkWidget *box;
	PICTURE_LOAD(ui.icon,icon);
	gdk_window_set_icon(ui.mainwin->window,NULL,ui.icon.pixmap,NULL);

	PICTURE_LOAD(aboutpic,about);
	ui.aboutbox=CreatePixmapBox(_("About") ,&aboutpic,NULL    );
	ui.rulesbox=CreatePixmapBox(_("Rules") ,NULL     ,_(rulesstr));
	ui.storybox=CreatePixmapBox(_("Story") ,NULL     ,_(storystr));
	ui.pausebox=CreatePixmapBox(_("Paused"),&ui.icon ,_("Game paused. Press [Okay] to continue."));
	gtk_window_set_modal(GTK_WINDOW(ui.pausebox),TRUE);
	gtk_signal_connect_object(GTK_OBJECT(ui.pausebox),"show",GTK_SIGNAL_FUNC(ui_pause_game ),GTK_OBJECT(ui.pausebox));
/*	gtk_signal_connect_object(GTK_OBJECT(ui.pausebox),"hide",GTK_SIGNAL_FUNC(ui_resume_game),GTK_OBJECT(ui.pausebox));*/
	ui.scorebox=CreatePixmapBox(_("Score"),(Picture *)-1,"");
	modalquit(ui.scorebox);
	ui.scoretext=ui.lastCPBtext;
	gtk_label_set_justify(GTK_LABEL(ui.scoretext),GTK_JUSTIFY_LEFT);
	ui.endgamebox=CreatePixmapBox(_("Endgame"),NULL,_(endgamestr));
	modalquit(ui.endgamebox);
	ui.highscorebox=CreatePixmapBox(_("HighScore"),NULL,_("High Score:"));
	box=ui.lastCPBtext->parent;
		{ GtkWidget *clist;
static char *labs[3];
		labs[0]=_("Name"); labs[1]=_("Level"); labs[2]=_("Score");
		clist=gtk_clist_new_with_titles(3,labs);
		gtk_clist_set_selection_mode(GTK_CLIST(clist),GTK_SELECTION_SINGLE);
		gtk_clist_column_titles_passive(GTK_CLIST(clist));
		gtk_clist_set_column_justification(GTK_CLIST(clist),0,GTK_JUSTIFY_LEFT);
		gtk_clist_set_column_justification(GTK_CLIST(clist),1,GTK_JUSTIFY_RIGHT);
		gtk_clist_set_column_justification(GTK_CLIST(clist),2,GTK_JUSTIFY_RIGHT);

		/* Make titles centered, if possible */
#ifdef HAVE_GTK_CLIST_GET_COLUMN_WIDGET
			{ int i;
			for (i=0;i<3;i++)
				gtk_alignment_set(GTK_ALIGNMENT(gtk_clist_get_column_widget(GTK_CLIST(clist),i)),0.5,0.5,0.0,0.0);
			}
#endif

		gtk_clist_set_column_width(GTK_CLIST(clist),0,180);
		gtk_clist_set_column_width(GTK_CLIST(clist),1,35);
		gtk_clist_set_column_width(GTK_CLIST(clist),2,60);
#ifdef HAVE_GTK_CLIST_SET_POLICY
		gtk_clist_set_policy(GTK_CLIST(clist),GTK_POLICY_AUTOMATIC,GTK_POLICY_AUTOMATIC);
		/* We are probably running on pre-1.1 GTK+ */
		gtk_widget_set_usize(clist,(180+35+60)*5/4,250);
#endif
		gtk_box_pack_start_defaults(GTK_BOX(box),clist);
		gtk_widget_show(clist);
		ui.highscorelist=clist;
		}
	ui.enternameentry=newentry(_("You earned a high score.\nEnter your name:"));
	ui.enternamebox=newentrybox(_("Enter Name"),_("OK"),ui.enternameentry,&box,GTK_SIGNAL_FUNC(enternamefunc));
	gtk_box_pack_start_defaults(GTK_BOX(box),ui.enternameentry->parent);

	modalquit(ui.enternamebox);
}

void ui_popup_dialog(int dialog) {
	switch (dialog) {
		case game_HIGHSCORE: gtk_widget_show(ui.highscorebox); return;
		case game_ENTERNAME: gtk_widget_show(ui.enternamebox); break;
		case game_SCORE    : gtk_widget_show(ui.    scorebox); break;
		case game_ENDGAME  : gtk_widget_show(ui.  endgamebox); break;
	}
	gtk_main();
}

/*********************/
/* Graphics routines */
/*********************/

void ui_set_cursor(int cursor) {
	switch (cursor) {
	case game_BUCKETC:
		mcursor_set(&bucket.cursor); break;
	case game_DOWNC:
		mcursor_set(&ui.downcursor); break;
	case game_DEFAULTC:
		mcursor_set(&ui.defaultcursor); break;
	default:
		mcursor_set(&OS.cursor[cursor]); break;
	}
}

void ui_clear(void) {
	if (!ui.whitegc) return;
	gdk_draw_rectangle(ui.offscreen,
		ui.whitegc,
		TRUE,
		0, 0,
		game_scrwidth,game_scrheight);
}

void ui_refresh(void)
{
	gtk_widget_draw(ui.drawarea,NULL);
	gdk_flush();
}

void ui_draw(Picture *pic,int x,int y)
{
	gdk_gc_set_clip_origin(pic->gc,x,y);
	gdk_draw_pixmap(ui.offscreen,pic->gc,pic->pixmap,0,0,x,y,pic->width,pic->height);
}

/* X is in WIDTH/2, Y is in 0 */
void ui_draw_string(const char *string,int x,int y,GdkGC *gc)
{
#define BORDER 2
GdkFont *font=ui.mainwinstyle->font;
gint width,height;

/* lbearing&rbearing - what are they for? */
	width=gdk_string_width(font,string);
	height=font->ascent+font->descent;
	gdk_draw_rectangle(ui.offscreen,gc,TRUE,
		x-width/2-BORDER-1,y  ,width+2*(BORDER+1)-1,height+2*(BORDER+1)-1);
	gdk_draw_rectangle(ui.offscreen,ui.mainwinstyle->black_gc,FALSE,
		x-width/2-BORDER-1,y  ,width+2*(BORDER+1)-1,height+2*(BORDER+1)-1);
	gdk_draw_string(ui.offscreen,font,ui.linegc,x-width/2,y+font->ascent+BORDER,string);
#undef BORDER
}

void ui_draw_centered (Picture *pic)
{ ui_draw(pic,(game_scrwidth-pic->width)/2,(game_scrheight-pic->height)/2); }

void ui_draw_line(int x1, int y1, int x2, int y2)
{
	if (!ui.linegc) return;
	gdk_draw_line(ui.offscreen,ui.linegc,x1,y1,x2,y2);
}

/******************/
/* Other routines */
/******************/

static char *ghost_over;
void ui_set_pausebutton (int action)
{
GtkWidget *w;
#ifdef HAVE_GNOME
	w=menu_game[PAUSE_MENU_NUM].widget;
#else /* HAVE_GNOME */
char *s=(ghost_over?ghost_over:_("<Main>/Game/Pause game"));
#ifdef HAVE_GTK_ITEM_FACTORY_GET_WIDGET
	w=gtk_item_factory_get_widget(ui.mainwinfactory,s);
#else /* !HAVE_GNOME && HAVE_GTK_ITEM_FACTORY_GET_WIDGET */
GtkMenuPath *mp=gtk_menu_factory_find(menufactory10,s);
	w=(mp?mp->widget:NULL);
#endif /* !HAVE_GNOME && HAVE_GTK_ITEM_FACTORY_GET_WIDGET */
#endif /* HAVE_GNOME */
	if (w) gtk_widget_set_sensitive(w,action);
}

#ifdef HAVE_LIBGGN
void ui_set_ggndisconnbutton (int action)
{
	ghost_over=_("<Main>/Game/Disconnect...");
	ui_set_pausebutton(action);
	ghost_over=NULL;
}
#endif

void ui_MainLoop(void) { gtk_main(); }

void ui_update_hsbox(void) {
int i;
char *vals[3];
char buf[2][20];
	gtk_clist_freeze(GTK_CLIST(ui.highscorelist));
	gtk_clist_clear (GTK_CLIST(ui.highscorelist));
	vals[1]=buf[0]; vals[2]=buf[1];
	for (i=0;i<10;i++) {
		vals[0]=scores.name[i];
		VARPRINTF(buf[0],"%d",scores.level[i]);
		VARPRINTF(buf[1],"%d",scores.score[i]);
		gtk_clist_append(GTK_CLIST(ui.highscorelist),vals);
		}
	gtk_clist_thaw  (GTK_CLIST(ui.highscorelist));
}

void ui_update_scorebox(int level, int score) {
	char str[40];
	sprintf (str, _("After Level %d.\nYour score: %d"), level, score);
	gtk_label_set_text(GTK_LABEL(ui.scoretext),str);
}
